/* acapd.c - acap server entry point
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define ACAPD_C
#include <config.h>
#include <stdlib.h>
#include <unistd.h>
#include <com_err.h>
#include <limits.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <db.h>

#ifdef HAVE_SCHED_H
#include <sched.h>
#endif				/* HAVE_SCHED_H */

#ifdef __sun			/* TODO: fix this bogosity */
#include <thread.h>
#endif

#include "acap_err.h"
#include "client.h"
#include "../daemon/sock.h"
#include "../daemon/daemon.h"
#include "../daemon/conn.h"
#include "../daemon/daemon_err.h"
#include "command.h"
#include "dataset_ref.h"
#include "dataset_data.h"

struct __db_env *acapd_db_env;

/* I  know, I shouldn't need explicit initializers,
 * but they make me Happy. */
int             debug_all = 0,
                debug_acapd_io = 0,
                debug_daemon_support = 0,
                debug_daemon_rselock = 0,
                debug_daemon_io = 0,
                debug_acapd_interlock = 0,
                debug_acapd_search = 0,
                debug_acapd_dataset = 0,
                debug_acapd_store = 0,
                debug_acapd_data = 0,
                debug_acapd_context = 0,
                debug_acapd_client = 0,
                debug_acapd_dataset_db = 0;

static int      emacs_debugging_info = 0;

static void
                log_db_err(const char *db_errpfx __attribute__((unused)),
			                   char *buffer) {
  fprintf(stderr, "DB: %s\n", buffer);
}

static sock_t  *acap_sock = NULL;

static void
terminate(int sig __attribute__((unused)))
{
  if (acap_sock)
    sock_close(&acap_sock);
  client_shutdown();
  command_shutdown();
}

static          result_t
setup_terminate(sigset_t * oldmask)
{
  static const int terminate_sigs[] = {
    SIGHUP,
    SIGINT,
    SIGQUIT,
    SIGTERM,
    0
  };

  const int      *sig;

  struct sigaction sa;

  sigset_t        mask;

  memset(&sa, 0L, sizeof(sa));
  sa.sa_handler = &terminate;
  sigemptyset(&sa.sa_mask);
  sigemptyset(&mask);
  for (sig = terminate_sigs; *sig; sig++)
    sigaddset(&mask, *sig);
  sigprocmask(SIG_BLOCK, &mask, oldmask);
  for (sig = terminate_sigs; *sig; sig++)
    if (sigaction(*sig, &sa, NULL))
      return errno;

  return OK;
}

int
main(int argc, char **argv)
{
  result_t        result = OK;

  int             c;

  const char     *port = ACAP_DEFAULT_PORT_NAME;

  int             errflg = 0;

  char           *options,
                 *value;

  DB_ENV          db_env;

  DB_TXN_STAT    *db_txn_stats;

  bool_t          inited_db = FALSE;

  const char     *acapd_root = ACAPD_ROOT;

  sigset_t        oldmask;

#if 0
  char           *db_config[] = {
    "DB_DATA_DIR data",
    "DB_LOG_DIR log",
    NULL
  };

#endif

  {
    /*
     * We use this for randomness, not crypto, so there's no reason to do
     * anything fancy here...
     */
    struct timeval  t;

    gettimeofday(&t, NULL);
    srand48(t.tv_sec ^ t.tv_usec);
  }

  initialize_acpd_error_table();

  while ((c = getopt(argc, argv, "eh:p:d:t:c:")) != EOF)
    switch (c) {
    case 'e':
      emacs_debugging_info = 1;
      break;
    case 'h':
      acapd_root = optarg;
      break;
    case 'p':
      port = optarg;
      break;
    case 'd':
      {
	const char     *const debugopts[] = {
#define DEBUG_ALL 0
	  "all",
#define DEBUG_ACAPD 1
	  "acapd",
#define DEBUG_ACAPD_IO 2
	  "acapd_io",
#define DEBUG_ACAPD_PARSER 3
	  "acapd_parser",
#define DEBUG_DAEMON_SUPPORT 4
	  "daemon_support",
#define DEBUG_DAEMON_RSELOCK 5
	  "daemon_rselock",
#define DEBUG_DAEMON_IO 6
	  "daemon_io",
#define DEBUG_ACAPD_INTERLOCK 7
	  "acapd_interlock",
#define DEBUG_ACAPD_SEARCH 8
	  "acapd_search",
#define DEBUG_ACAPD_DATASET 9
	  "acapd_dataset",
#define DEBUG_ACAPD_STORE 10
	  "acapd_store",
#define DEBUG_ACAPD_DATA 11
	  "acapd_data",
#define DEBUG_ACAPD_CONTEXT 12
	  "acapd_context",
#define DEBUG_ACAPD_CLIENT 13
	  "acapd_client",
#define DEBUG_ACAPD_DATASET_DB 14
	  "acapd_dataset_db",
	  NULL
	};

	options = optarg;
	while (*options != '\0')
	  switch (getsubopt(&options, debugopts, &value)) {
	  case DEBUG_ALL:
	    debug_all
	      = debug_acapd_io
	      = acap_debug
	      = debug_daemon_support
	      = debug_daemon_rselock
	      = debug_daemon_io
	      = debug_acapd_interlock
	      = debug_acapd_search
	      = debug_acapd_dataset
	      = debug_acapd_store
	      = debug_acapd_data
	      = debug_acapd_context
	      = debug_acapd_client
	      = debug_acapd_dataset_db
	      = 1;
	    break;
	  case DEBUG_ACAPD:
	    debug_acapd_io
	      = acap_debug
	      = debug_acapd_interlock
	      = debug_acapd_search
	      = debug_acapd_dataset
	      = debug_acapd_store
	      = debug_acapd_data
	      = debug_acapd_context
	      = debug_acapd_client
	      = debug_acapd_dataset_db
	      = 1;
	    break;
	  case DEBUG_ACAPD_IO:
	    debug_acapd_io = 1;
	    break;
	  case DEBUG_ACAPD_PARSER:
	    acap_debug = 1;
	    break;
	  case DEBUG_DAEMON_SUPPORT:
	    debug_daemon_support = 1;
	    break;
	  case DEBUG_DAEMON_RSELOCK:
	    debug_daemon_rselock = 1;
	    break;
	  case DEBUG_DAEMON_IO:
	    debug_daemon_io = 1;
	    break;
	  case DEBUG_ACAPD_INTERLOCK:
	    debug_acapd_interlock = 1;
	    break;
	  case DEBUG_ACAPD_SEARCH:
	    debug_acapd_search = 1;
	    break;
	  case DEBUG_ACAPD_DATASET:
	    debug_acapd_dataset = 1;
	    break;
	  case DEBUG_ACAPD_STORE:
	    debug_acapd_store = 1;
	    break;
	  case DEBUG_ACAPD_DATA:
	    debug_acapd_data = 1;
	    break;
	  case DEBUG_ACAPD_CONTEXT:
	    debug_acapd_context = 1;
	    break;
	  case DEBUG_ACAPD_CLIENT:
	    debug_acapd_client = 1;
	    break;
	  case DEBUG_ACAPD_DATASET_DB:
	    debug_acapd_dataset_db = 1;
	    break;
	  default:
	    errflg = 1;
	  }
	break;
      }
    case 't':
      {
	const char     *const concurrentopts[] = {
#define CONCURRENT_CONN 0
	  "conn",
#define CONCURRENT_COMMAND 1
	  "command",
	  NULL
	};

	options = optarg;
	while (*options != '\0') {
	  int             which;

	  unsigned        concurrency;

	  char           *end;

	  which = getsubopt(&options, concurrentopts, &value);
	  if (value)
	    concurrency = strtoul(value, &end, 10);
	  else
	    concurrency = 0;
	  if (*end || !concurrency || concurrency == ULONG_MAX)
	    errflg = 1;
	  else
	    switch (which) {
	    case CONCURRENT_CONN:
	      conn_set_concurrency(concurrency);
	      break;
	    case CONCURRENT_COMMAND:
	      command_set_concurrency(concurrency);
	      break;
	    default:
	      errflg = 1;
	    }
	}
	break;
      }
    case 'c':
      {
	const char     *const cacheopts[] = {
#define CACHE_DATASET 0
	  "dataset",
#define CACHE_DATA 1
	  "data",
	  NULL
	};

	options = optarg;
	while (*options != '\0') {
	  int             which;

	  unsigned        cachesize;

	  char           *end;

	  which = getsubopt(&options, cacheopts, &value);
	  if (value)
	    cachesize = strtoul(value, &end, 10);
	  else
	    cachesize = 0;
	  if (*end || cachesize == ULONG_MAX)
	    errflg = 1;
	  else
	    switch (which) {
	    case CACHE_DATASET:
	      dataset_ref_set_cachesize(cachesize);
	      break;
	    case CACHE_DATA:
	      dataset_data_set_cachesize(cachesize);
	      break;
	    default:
	      errflg = 1;
	    }
	}
	break;
      }
    default:
      errflg = 1;
    }

  if (errflg) {
    com_err(argv[0], ACAP_ERR_USAGE, NULL);
    return 1;
  }
  result = daemon_init(argv[0]);
  if (result != OK)
    daemon_fatal(result, "Basic daemon initialization failed");

  result = setup_terminate(&oldmask);

  if (result != OK)
    daemon_fatal(result, "Unable to setup termination code");

  memset(&db_env, 0, sizeof(db_env));
  db_env.db_errcall = &log_db_err;
  db_env.db_verbose = 1;

#ifdef HAVE_THR_YIELD
  db_jump_set((void *) &thr_yield, DB_FUNC_YIELD);
#else
  db_jump_set((void *) &sched_yield, DB_FUNC_YIELD);
#endif

  result = db_appinit(acapd_root, NULL, &db_env,
		      DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG
		      | DB_INIT_MPOOL | DB_INIT_TXN | DB_THREAD
		      | DB_USE_ENVIRON);
  if (result == DB_RECOVER)
    daemon_fatal(ACAP_ERR_DB_RECOVER_NEEDED, "Try running db_recover -h %s\n", acapd_root);

  if (result == DB_RECOVER_FATAL)
    daemon_fatal(ACAP_ERR_DB_RECOVER_FATAL_NEEDED, "Try running db_recover -c -h %s\n", acapd_root);

  if (result < 0)
    result = ACAP_ERR_DB_INIT;

  if (result != OK)
    daemon_fatal(result, "Database initialization failed");

  inited_db = TRUE;

  /*
   * Just for laughs, let's see if there're any outstanding transactions, and
   * if there are, let's blow them away...
   */
  result = txn_stat(db_env.tx_info, &db_txn_stats, &malloc);
  if (result < 0)
    result = ACAP_ERR_DB;
  if (result != OK)
    daemon_fatal(result, "Failed to retrieve daemon transaction status");
  if (db_txn_stats->st_nactive != 0) {
    /*
     * Grn -- there're active transactions.  Well, they're not ours... which
     * means that something bad's happening.
     */
    db_appexit(&db_env);
    daemon_fatal(ACAP_ERR_DB_TRX_ACTIVE, "Try running db_recover -h %s\n", acapd_root);
  }
  free(db_txn_stats);

  acapd_db_env = &db_env;

  result = sock_open(port, ACAP_DEFAULT_PORT_NUMBER,
		     &client_start, NULL, &acap_sock);
  if (result != OK)
    daemon_fatal(result, "Unable to listen on the specified socket (%s)", port);

  /* Allow termination signals... */
  sigprocmask(SIG_SETMASK, &oldmask, NULL);

  VL(("Starting ACAP daemon in %s\n", acapd_root));
  result = daemon_run();
  if (result != OK)
    daemon_fatal(result, "while running");

  daemon_shutdown();

  if (inited_db)
    db_appexit(&db_env);

  return (result != OK) ? EXIT_FAILURE : EXIT_SUCCESS;
}

void
acapd_log(FILE * file,
	  const char *format,
	  ...)
{
  va_list         ap;

  va_start(ap, format);
  fprintf(file, "%lu: ", (unsigned long) pthread_self());
  vfprintf(file, format, ap);
  va_end(ap);
}

void
acapd_failed_assertion(const char *expression,
		       const char *file,
		       unsigned line)
{
  if (emacs_debugging_info)
    daemon_fatal(DAEMON_ERR_FAILED_ASSERTION,
		 "%s\n    at %s:%d\n%s:%d:\n",
		 expression, file, line, file, line);
  else
    daemon_fatal(DAEMON_ERR_FAILED_ASSERTION,
		 "%s\n    at %s:%d\n",
		 expression, file, line);
}
