/* client.c - client implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define CLIENT_C
#include <config.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <com_err.h>
#include <stdarg.h>
#include "client.h"
#include "respcode.h"
#include "token.acap.h"
#include "../daemon/daemon.h"
#include "../daemon/daemon_err.h"
#include "../util/util_err.h"
#include "acap_err.h"
#include "context.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_acapd_client)

extern int      acap_parse(client_t * client);

static str_t   *anon_username = (str_t *) "\0\0\0\011anonymous";

static pthread_key_t current_client_key;

static pthread_once_t current_client_key_once = PTHREAD_ONCE_INIT;

static pthread_mutex_t client_set_mutex = PTHREAD_MUTEX_INITIALIZER;

static hash_table_t client_set = HASH_INITIALIZER(&hash_ptr_procs);

static void
current_client_key_init()
{
  result_t result = pthread_key_create(&current_client_key, NULL);
  if (result)
    daemon_fatal(result, "fatal error: pthread_key_create");
}

static result_t
client_free(client_t ** client)
{
  result_t        result;

  hash_entry_t   *hash_entry;

  I(client);
  I(*client);

  VL(("Freeing client %p\n", *client));

  pthread_mutex_lock(&client_set_mutex);
  result = hash_find(&client_set, *client, HASH_NO_INSERT, NULL, &hash_entry);
  I(result != HASH_NOT_FOUND);
  hash_delete_entry(hash_entry);
  pthread_mutex_unlock(&client_set_mutex);

  conn_flush(client_CONN(*client));

  refcount_destroy(&(*client)->refcount);
  conn_free(&client_CONN(*client));
  pthread_mutex_destroy(client_OUTPUT_MUTEX(*client));
  pthread_mutex_destroy(&(*client)->contexts_mutex);
  hash_destroy(&(*client)->contexts);
  sasl_dispose(&(*client)->sasl);
  if ((*client)->username != anon_username)
    str_free(&(*client)->username);

  free(*client);
  *client = NULL;

  return OK;
}

result_t
client_ref(client_t * client)
{
  VL(("Adding a ref to client %p\n", client));

  return refcount_REF(&client->refcount);
}

result_t
client_unref(client_t ** client)
{
  result_t        result;

  VL(("Removing a ref from client %p\n", *client));

  result = refcount_UNREF(&(*client)->refcount,
			  client_free(client));

  *client = NULL;		/* either way... */
  return result;
}

result_t
client_lookup_context(client_t * client,
		      const str_t * name,
		      struct context_s ** context)
{
  result_t        result;

  hash_entry_t   *entry;

  I(client);
  I(name);
  I(context);

  pthread_mutex_lock(&client->contexts_mutex);

  VL(("Looking up context for %s in client %p\n",
      str_DATA(name), client));

  result = hash_find(&client->contexts, name, HASH_NO_INSERT, NULL, &entry);
  if (result == HASH_NOT_FOUND)
    result = ACAP_ERR_NO_SUCH_CONTEXT;

  if (result == OK) {
    *context = (context_t *) hash_GET_VALUE(entry);
    context_ref(*context);
  } else
    *context = NULL;

  VL(("Found context %p\n", *context));

  pthread_mutex_unlock(&client->contexts_mutex);

  return result;
}

result_t
client_map_context(client_t * client,
		   const str_t * name,
		   struct context_s * context,
		   bool_t replace_previous)
{
  result_t        result = OK;

  hash_entry_t   *entry;

  hash_did_insert_t inserted;

  I(client);
  I(name);
  I(context);

  pthread_mutex_lock(&client->contexts_mutex);

  VL(("Mapping context %s to %p\n",
      str_DATA(name), context));

  if (client->forgetting_contexts)
    goto cleanup;

  result = hash_find(&client->contexts, name, HASH_INSERT,
		     &inserted, &entry);

  if (result != OK)
    goto cleanup;

  I(entry);

  if (inserted == HASH_DID_NOT_INSERT) {
    context_t      *old;

    old = (context_t *) hash_GET_VALUE(entry);
    if (replace_previous)
      context_set_previous(context, hash_GET_KEY(entry), old);
    else
      context_unref(&old);
  }
  context_ref(context);
  hash_SET_VALUE(entry, context);

cleanup:
  pthread_mutex_unlock(&client->contexts_mutex);

  return result;
}

result_t
client_forget_context(client_t * client,
		      const str_t * name)
{
  result_t        result;

  hash_entry_t   *entry;

  context_t      *context;

  I(client);
  I(name);

  pthread_mutex_lock(&client->contexts_mutex);

  VL(("Forgetting context %s for client %p\n",
      str_DATA(name), client));

  result = hash_find(&client->contexts, name, HASH_NO_INSERT, NULL, &entry);
  if (result == HASH_NOT_FOUND) {
    result = ACAP_ERR_NO_SUCH_CONTEXT;
    goto cleanup;
  }
  I(entry);

  context = (context_t *) hash_GET_VALUE(entry);
  I(context);

  context_unref(&context);

cleanup:
  pthread_mutex_unlock(&client->contexts_mutex);
  return result;
}

static void
client_forget_contexts(client_t * client)
{
  hash_iter_t     it;

  hash_entry_t   *entry;

  I(client);

  pthread_mutex_lock(&client->contexts_mutex);

  VL(("Forgetting contexts for client %p\n", client));

  client->forgetting_contexts = TRUE;

  hash_iter_begin(&client->contexts, &it);
  while (hash_iter_next(&it, &entry) == OK) {
    context_t      *context;

    context = (context_t *) hash_GET_VALUE(entry);
    context_unref(&context);
  }

  VL(("Done forgetting contexts for client %p\n", client));

  pthread_mutex_unlock(&client->contexts_mutex);
}

static void
client_parse(client_t * client, bool_t * connection_closed)
{
  int             closing;

  I(client);
  I(connection_closed);

  VL(("Parsing data from client %p\n", client));

  pthread_once(&current_client_key_once,
	       &current_client_key_init);

  pthread_setspecific(current_client_key, client);
  closing = acap_parse(client);
  pthread_setspecific(current_client_key, NULL);

  if (closing) {
    client_forget_contexts(client);
    client_unref(&client);
    *connection_closed = TRUE;
  } else {
    *connection_closed = FALSE;
  }
}

/* XXX This should be const */
static sasl_callback_t sasl_callbacks[] = {
  {
    SASL_CB_LIST_END, NULL, NULL
  }
};

static result_t
client_alloc(client_t ** client,
	     struct sockaddr_in * peer,
	     int fd)
{
  result_t        result = OK;

  int             sasl_result;

  char           *init_str = NULL;

  size_t          init_str_len;


  I(client);
  I(!*client);
  I(peer);
  I(0 <= fd);

  *client = malloc(sizeof(client_t));
  if (!*client)
    return errno;

  result = pthread_mutex_init(client_OUTPUT_MUTEX(*client), NULL);
  if (result)
    goto cleanup;

  result = pthread_mutex_init(&(*client)->contexts_mutex, NULL);
  if (result)
    goto cleanup_output_mutex;

  (*client)->forgetting_contexts = FALSE;

  result = hash_init(&hash_str_procs,
		     &(*client)->contexts);
  if (result)
    goto cleanup_contexts_mutex;

  result = refcount_init(&(*client)->refcount);
  if (result)
    goto cleanup_contexts_hash;

  client_ref(*client);		/* goes away when conn goes away */

  client_SET_STATE(*client, CLIENT_TAG);

  (*client)->current_tag = NULL;
  (*client)->authenticated = FALSE;
  (*client)->recovering = FALSE;
  (*client)->isbad = FALSE;
  (*client)->parser_cleanup = NULL;

  (*client)->username = anon_username;

  (*client)->sasl = NULL;
  sasl_result = sasl_server_new("acap", NULL, NULL, sasl_callbacks,
				SASL_SECURITY_LAYER, &(*client)->sasl);
  if (sasl_result != SASL_OK) {
    result = DAEMON_ERR_SASL;
    goto cleanup_refcount;
  }
  sasl_result = sasl_setprop((*client)->sasl, SASL_IP_REMOTE, peer);

  if (sasl_result == SASL_OK) {
    struct sockaddr_in sin;

    int             len = sizeof(sin);

    if (getsockname(fd, (struct sockaddr *) & sin, &len)) {
      result = errno;
      goto cleanup_sasl;
    }
    sasl_result = sasl_setprop((*client)->sasl, SASL_IP_LOCAL, &sin);
  }
  if (sasl_result != SASL_OK) {
    result = DAEMON_ERR_SASL;
    goto cleanup_sasl;
  }
  pthread_mutex_lock(client_OUTPUT_MUTEX(*client));

  (*client)->conn = NULL;
  result = conn_alloc(&(*client)->conn, peer, fd,
		      (conn_parser_t *) & client_parse,
		      *client);
  if (result)
    goto cleanup_mutex_lock;

  sasl_result = sasl_listmech((*client)->sasl, NULL,
			      "* ACAP (ContextLimit \"100\") "
			      "(Implementation "
			      "\"Carnegie Mellon " PACKAGE "-" VERSION "\") "
			      "(SASL \"",
			      "\" \"",
			      "\")\r\n",
			      &init_str,
			      &init_str_len,
			      NULL);
  if (sasl_result != SASL_OK)
    result = DAEMON_ERR_SASL;

  else if (write(conn_FD(client_CONN(*client)),
		 init_str,
		 init_str_len) != (ssize_t) init_str_len)
    result = errno;

  if (init_str)
    free(init_str);

  if (result)
    goto cleanup_mutex_lock;

  pthread_mutex_unlock(client_OUTPUT_MUTEX(*client));

  {
    hash_did_insert_t inserted;

    hash_entry_t   *hash_entry;

    pthread_mutex_lock(&client_set_mutex);
    result = hash_find(&client_set, *client, HASH_INSERT, &inserted, &hash_entry);
    if (result != OK)
      goto cleanup_conn;
    I(inserted == HASH_DID_INSERT);
    pthread_mutex_unlock(&client_set_mutex);
  }

  return OK;

cleanup_mutex_lock:
  pthread_mutex_unlock(client_OUTPUT_MUTEX(*client));

cleanup_conn:
  if (result) {
    /* Attempt to tell the other side that we're dying */
    const char     *errstr;

    if (result == DAEMON_ERR_SASL) {
      /* Error was in SASL */
      errstr = sasl_errstring(sasl_result, NULL, NULL);
    } else {
      /* Error was something else - use error_message */
      errstr = error_message(result);
    }
    client_printf(*client, "* Bye %z\r\n", errstr);
    conn_flush(client_CONN(*client));
  }
  conn_free(&(*client)->conn);

cleanup_sasl:
  sasl_dispose(&(*client)->sasl);

cleanup_refcount:
  refcount_destroy(&(*client)->refcount);

cleanup_contexts_hash:
  hash_destroy(&(*client)->contexts);

cleanup_contexts_mutex:
  pthread_mutex_destroy(&(*client)->contexts_mutex);

cleanup_output_mutex:
  pthread_mutex_destroy(client_OUTPUT_MUTEX(*client));

cleanup:
  free(*client);
  *client = NULL;

  return result;
}

result_t
client_start(struct sockaddr_in * peer,
	     int fd,
	     void *rock __attribute__((unused)))
{
  result_t        result;

  client_t       *client = NULL;

  I(peer);
  I(0 <= fd);

  result = client_alloc(&client, peer, fd);
  if (result)
    return result;

  VL(("Allocated client handler %p for fd %d\n", client, fd));

  return OK;
}

result_t
client_current(client_t ** client)
{
  I(client);
  I(!*client);

  *client = pthread_getspecific(current_client_key);
  if (!*client)
    return ACAP_ERR_CLIENT_NOT_SET;

  return OK;
}

static inline   bool_t
safe_to_use_quoted(const char *it,
		   const char *end)
{
  unsigned        len;

  I(it);
  I(end);
  I(it <= end);

  len = end - it;

  if (len > ACAP_MAX_QSTR_LEN)
    return FALSE;

  while (it < end) {
    if (*it == '\0'		/* check illegal chars... */
	|| *it == '\r'
	|| *it == '\n'

#ifdef __CHAR_UNSIGNED__
	|| 0x7F < *it
#else
	|| *it < 0
#endif

      )
      return FALSE;
    if (*it == '\"'		/* check len, with \ escapes... */
	|| *it == '\\')
      if (ACAP_MAX_QSTR_LEN < ++len)
	return FALSE;
    ++it;
  }
  return TRUE;
}

static result_t
client_print_uint(client_t * client,
		  uint32_t number)
{
  char            buf[10],	/* big enough for any 32bit integer */
                 *start,
                 *here,
                 *end;

  uint32_t        sub = number;

  start = here = buf;

  /* First, we write it, reverse order... */
  do {
    *here++ = sub % 10 + '0';
    sub /= 10;
  } while (sub);

  end = here;			/* save the end */

  here--;			/* back to last char written */

  /* Now we need to reverse start through here... */
  while (start < here) {
    char            a = *start;

    *start = *here;
    *here = a;
    ++start;
    --here;
  }
  /* And now we can write it. */
  return conn_write(client_CONN(client), buf, end - buf);
}

static result_t
client_print_qstr(client_t * client,
		  const char *data,
		  unsigned len)
{
  result_t        result;

  I(client);
  I(data);

  /* Okay - what type of string will we send? */
  if (safe_to_use_quoted(data, data + len)) {
    /* A quoted string. */
    /* Initial quote... */
    result = conn_putc(client_CONN(client), '\"');
    if (result)
      return result;

    /* Filler... */
    while (len) {
      if (*data == '\\' || *data == '\"') {
	result = conn_putc(client_CONN(client), '\\');
	if (result)
	  return result;
      }
      result = conn_putc(client_CONN(client), *data);
      if (result)
	return result;

      ++data;
      --len;
    }

    /* Trailing quote... */
    result = conn_putc(client_CONN(client), '\"');
    if (result)
      return result;
  } else {
    /* Oh, goody - a length-string. */
    result = conn_putc(client_CONN(client), '{');
    if (result)
      return result;

    result = client_print_uint(client, len);
    if (result)
      return result;

    result = conn_putc(client_CONN(client), '}');
    if (result)
      return result;

    result = conn_putc(client_CONN(client), '\r');
    if (result)
      return result;

    result = conn_putc(client_CONN(client), '\n');
    if (result)
      return result;

    result = conn_write(client_CONN(client), data, len);
    if (result)
      return result;
  }

  return OK;
}

result_t
client_printf(client_t * client, const char *format,...)
{
  result_t        result = OK;

  va_list         va;

  I(client);
  I(format);

  va_start(va, format);

  for (; *format && !result; format++)
    if (*format != '%') {
      result = conn_putc(client_CONN(client), *format);
    } else {
      format++;
      switch (*format) {
      case 't':
	{
	  /* unquoted null-terminated string */
	  const char     *str = va_arg(va, const char *);

	  I(str);
	  result = conn_write(client_CONN(client), str, strlen(str));
	  break;
	}
      case 's':
	{
	  /* quoted str_t */
	  const str_t    *str = va_arg(va, str_t *);

	  I(str);
	  result = client_print_qstr(client, str_DATA(str), str_LEN(str));
	  break;
	}
      case 'z':
	{
	  /* quoted null-terminated string */
	  const char     *str = va_arg(va, const char *);

	  I(str);
	  result = client_print_qstr(client, str, strlen(str));
	  break;
	}
      case 'l':
	{
	  const char     *buffer = va_arg(va, const char *);

	  unsigned        len = va_arg(va, unsigned);

	  I(buffer);
	  result = client_print_qstr(client, buffer, len);
	  break;
	}
      case 'u':
	{
	  uint32_t        number = va_arg(va, uint32_t);

	  result = client_print_uint(client, number);
	  break;
	}
      default:
	daemon_fatal(DAEMON_ERR_INTERNAL, "Illegal format");
      }
    }

  va_end(va);
  return result;
}

result_t
client_send_result(client_t * client,
		   str_t * tag,
		   result_t command_result,
		   const char *ok_message,
		   respcode_t * respcode)
{
  result_t        result;

  I(client);
  I(tag);
  I(ok_message);

  pthread_mutex_lock(client_OUTPUT_MUTEX(client));

  result = client_printf(client, "%t %t",
			 str_DATA(tag),
			 command_result == OK ? "Ok" : "No");

  if (result == OK && respcode)
    result = respcode_write(respcode, client, &command_result);
  if (respcode)
    respcode_free(&respcode);

  if (result == OK)
    result = client_printf(client, " %z\r\n",
			   command_result == OK ? ok_message
			   : error_message(command_result));

  conn_flush(client_CONN(client));

  pthread_mutex_unlock(client_OUTPUT_MUTEX(client));

  return result;
}

result_t
client_parser_add_cleanup(client_t * client,
			  client_parser_cleanup_proc_t * proc,
			  void *rock)
{
  result_t        result;

  client_parser_cleanup_t *cleanup;

  I(client);
  I(proc);

  cleanup = (client_parser_cleanup_t *)
    malloc(sizeof(client_parser_cleanup_t));
  if (!cleanup)
    return errno;

  cleanup->proc = proc;
  cleanup->rock = rock;

  result = sllist_alloc(cleanup,
			client->parser_cleanup,
			&client->parser_cleanup);
  if (result) {
    free(cleanup);
    return result;
  }
  return OK;
}

static result_t
client_parser_free_cleanup(sllist_t * cleanup,
			   void *rock __attribute__((unused)))
{
  I(cleanup);
  I(sllist_DATA(cleanup));

  free(sllist_DATA(cleanup));

  return OK;
}

result_t
client_parser_forget_cleanup(client_t * client)
{
  result_t        result;

  I(client);

  client_SET_TAG(client, NULL);

  result = sllist_apply(client->parser_cleanup,
			client_parser_free_cleanup, NULL);
  sllist_free(&(client->parser_cleanup));
  return result;

}

static result_t
client_parser_do_one_cleanup(sllist_t * _cleanup,
			     void *rock __attribute__((unused)))
{
  client_parser_cleanup_t *cleanup;

  I(_cleanup);
  I(sllist_DATA(_cleanup));

  cleanup = (client_parser_cleanup_t *) sllist_DATA(_cleanup);

  I(cleanup->proc);
  cleanup->proc(cleanup->rock);
  return OK;
}

result_t
client_parser_do_cleanup(client_t * client)
{
  I(client);

  client_SET_TAG(client, NULL);

  sllist_apply(client->parser_cleanup,
	       &client_parser_do_one_cleanup,
	       NULL);
  return client_parser_forget_cleanup(client);
}

void
client_shutdown(void)
{
  hash_iter_t     it;

  hash_entry_t   *entry;

  client_t       *client;

  pthread_mutex_lock(&client_set_mutex);
  hash_iter_begin(&client_set, &it);
  while (hash_iter_next(&it, &entry) != HASH_ITER_END) {
    client = *(client_t **) hash_GET_KEY(entry);
    pthread_mutex_lock(client_OUTPUT_MUTEX(client));
    client_printf(client, "* Bye \"Server shutting down\"\r\n");
    conn_flush(client_CONN(client));
    shutdown(conn_FD(client_CONN(client)), 2);
    pthread_mutex_unlock(client_OUTPUT_MUTEX(client));
  }
  pthread_mutex_unlock(&client_set_mutex);
}
