/* client.h - client interface
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#ifndef CLIENT_H
#define CLIENT_H

/* This is the ACAP-specific client stuff. */
#include "../daemon/conn.h"
#include "../util/str.h"
#include "../util/sllist.h"
#include "../util/refcount.h"
#include "../util/hash.h"

typedef enum {
  CLIENT_TAG,			/* we're expecting a tag */
  CLIENT_RECOVER,		/* recovering */
  CLIENT_RECOVER_CR,		/* recovering; we've seen a CR */
  CLIENT_CR,			/* we've seen a CR */
  CLIENT_QSTR,			/* processing a quoted string */
  CLIENT_LITERAL,		/* literals */
  CLIENT_NUMBER,		/* for parsing numbers */
  CLIENT_NORMAL,		/* normal state: parsing a line, expecting
				 * anything */
  CLIENT_ATOM			/* started processing an atom */
}               client_state_t;	/* current input state */

typedef void    client_parser_cleanup_proc_t(void *rock);

typedef struct {
  client_parser_cleanup_proc_t *proc;
  void           *rock;
}               client_parser_cleanup_t;

typedef struct client_s {
  refcount_t      refcount;	/* when can this client vanish? */
  conn_t         *conn;
  client_state_t  state;
  str_t          *current_tag;	/* tag of command currently being parsed */
  pthread_mutex_t output_mutex;
  pthread_mutex_t contexts_mutex;
  hash_table_t    contexts;
  bool_t          forgetting_contexts;
  sasl_conn_t    *sasl;
  str_t          *username;
  bool_t          authenticated;/* true after successful AUTHENTICATE */
  bool_t          recovering;	/* true iff the parser's recovering */
  bool_t          isbad;	/* whether the result bubbling up is BAD */
  sllist_t       *parser_cleanup;
}               client_t;

#define client_CONN(client) ((client)->conn)
#define client_SET_STATE(client, val) ((client)->state = (val))
#define client_SET_TAG(client, val) ((client)->current_tag = (val))
#define client_STATE(client) ((client)->state)
#define client_TAG(client) ((client)->current_tag)
#define client_OUTPUT_MUTEX(client) (&(client)->output_mutex)
#define client_SET_BAD(client) ((client)->isbad = TRUE)
#define client_IS_BAD(client) ((client)->isbad)
#define client_USERNAME(client) ((client)->username)

extern          result_t
client_start(struct sockaddr_in * peer,
	     int fd,
	     void *rock);

extern result_t client_current(client_t ** client);

/* client_printf is used to send data to clients.
 * The format is a specialized printf-style format;
 * the control characters are:
 * 	%t - unquoted null-terminated string (used for tags)
 *	%s - quoted, str_t
 *	%z - quoted, null-terminated string
 *	%l - quoted, buffer ptr followed by length
 *	%u - unsigned 32bit integer
 */
extern result_t
client_printf(client_t * client,
	      const char *format,
	      ...);

/* Just a utility function for command finalization...
 * automatically performs a flush.
 * This locks the client output mutex while writing.
 */

struct respcode_s;

extern result_t
client_send_result(client_t * client,
		   str_t * tag,
		   result_t result,
		   const char *ok_message,
		   struct respcode_s * respcode);

extern result_t
client_parser_add_cleanup(client_t * client,
			  client_parser_cleanup_proc_t * proc,
			  void *rock);

extern result_t client_parser_do_cleanup(client_t * client);

extern result_t client_parser_forget_cleanup(client_t * client);

/* For commands which need the client connection
 * to stick around... */
extern result_t client_ref(client_t * client);

extern result_t client_unref(client_t ** client);

/* Client context operations
 *
 * client_lookup_context() returns a referenced context; callers
 * should call context_unref() when finished.
 *
 * client_map_context() maps a name to a context, adding a reference
 * to it in the process.
 *
 * client_forget_context() revokes a name->context mapping and
 * dereferences the specified context, which will eventually cause
 * the context to go away.
 */
struct context_s;

extern result_t
client_lookup_context(client_t * client,
		      const str_t * name,
		      struct context_s ** context);

extern result_t
client_map_context(client_t * client,
		   const str_t * name,
		   struct context_s * context,
		   bool_t replace_previous);

extern result_t
client_forget_context(client_t * client,
		      const str_t * name);

/* Shuts down all current client connections */
extern void     client_shutdown(void);

#endif				/* CLIENT_H */
