/* dataset_db.c--master dataset db implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define DATASET_DB_C
#include <config.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/stat.h>
#include <string.h>
#include <db.h>
#include "dataset_db_fac.h"
#include "dataset_db.h"
#include "dataset_ref.h"
#include "acap_err.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_acapd_dataset_db)

result_t
dataset_db_alloc(hash_entry_t * hashent,
		 dataset_db_t ** dest)
{
  result_t        result;

  I(hashent);
  I(dest);

  *dest = (dataset_db_t *) malloc(sizeof(dataset_db_t));
  if (!*dest)
    return errno;

  (*dest)->hashent = hashent;

  VL(("Allocing dataset_db %p (\"%s\")\n",
      *dest, str_DATA(dataset_db_NAME(*dest))));

  result = refcount_init(dataset_db_REFCOUNT(*dest));
  if (result != OK)
    goto cleanup_dest;

  result = pthread_mutex_init(dataset_db_MUTEX(*dest), NULL);
  if (result != OK)
    goto cleanup_refcount;

  result = hash_init(&hash_uint32_procs,
		     dataset_db_DATASET_REFS(*dest));
  if (result != OK)
    goto cleanup_mutex;

  result = db_open(str_DATA(dataset_db_NAME(*dest)),
		   DB_HASH,
		   DB_CREATE | DB_THREAD,
		   S_IRUSR | S_IWUSR,
		   acapd_db_env,
		   NULL,
		   &(*dest)->db);
  if (result < 0)
    result = ACAP_ERR_DB_OPEN;

  if (result != OK)
    goto cleanup_hash;

  refcount_REF(&(*dest)->refcount);

  return OK;

cleanup_hash:
  hash_destroy(dataset_db_DATASET_REFS(*dest));

cleanup_mutex:
  pthread_mutex_destroy(dataset_db_MUTEX(*dest));

cleanup_refcount:
  refcount_destroy(dataset_db_REFCOUNT(*dest));

cleanup_dest:
  free(*dest);
  *dest = NULL;
  return result;
}

result_t
dataset_db_free(dataset_db_t ** db)
{
  I(db);
  I(*db);
  I((*db)->db);

  I(!hash_N_ENTRIES(dataset_db_DATASET_REFS(*db)));

  VL(("Freeing dataset_db %p (%s)\n", *db, dataset_db_NAME(*db)));

  refcount_destroy(dataset_db_REFCOUNT(*db));
  pthread_mutex_destroy(dataset_db_MUTEX(*db));
  hash_destroy(dataset_db_DATASET_REFS(*db));
  (*db)->db->close((*db)->db, 0);
  free(*db);
  *db = NULL;
  return OK;
}

result_t
dataset_db_ref(dataset_db_t * db)
{
  I(db);
  I(db->db);

  VL(("Adding a ref to dataset_db %p\n", db));

  return refcount_REF(&db->refcount);
}

result_t
dataset_db_unref(dataset_db_t ** db)
{
  result_t        result;

  I(db);
  I(*db);
  I((*db)->db);

  VL(("Removing a ref from dataset_db %p\n", *db));

  result = refcount_UNREF_CHECKLAST(dataset_db_REFCOUNT(*db),
				    dataset_db_fac_check_norefs(*db));
  *db = NULL;			/* either way... */

  return result;
}

result_t
dataset_db_lookup(dataset_db_t * db,
		  trx_t * trx,
		  dataset_key_t dataset_key,
		  dataset_ref_t ** dataset_ref)
{
  result_t        result;

  hash_did_insert_t inserted;

  hash_entry_t   *hashentry;

  DBT             key,
                  value;

  I(db);
  I(db->db);
  I(dataset_ref);

  *dataset_ref = NULL;

  VL(("Looking up key %u in dataset_db %p\n", dataset_key, db));

  pthread_mutex_lock(dataset_db_MUTEX(db));
  result = hash_find(dataset_db_DATASET_REFS(db),
		     (void *) dataset_key,
		     HASH_INSERT,
		     &inserted,
		     &hashentry);
  if (result != OK)
    goto cleanup_lock;

  /*
   * So, in one sense, we probably ought to create the dataset_ref_t, lock
   * it, fill in the hash table, release the lock on the dataset_db_t, read
   * in the dataset_t, fill in the dataset_ref_t, release it, and return it.
   * 
   * On the other hand, Berkeley DB has some of the most horrid locks I've ever
   * seen in my life--they're really, *really* bad.  I don't know why they
   * can't just take lock creation, destruction, lock, and unlock functions;
   * surely the overhead of a malloc is better than the overhead of the
   * spin/sleep() trick they do.  But the fact is, their locks suck, and
   * since it'd be using locks for parallel reads anyway (at least, that's
   * been my experience; I haven't traced through the code to be sure,
   * though), we might as well just hold onto a relatively zippy pthreads
   * lock while we read from the database.
   * 
   * So the easy thing is the right thing!  I love it when that happens.
   */

  if (inserted == HASH_DID_NOT_INSERT) {
    /* We found the dataset_ref_t, so we can just go ahead and use it. */
    *dataset_ref = (dataset_ref_t *) hash_GET_VALUE(hashentry);
    dataset_ref_ref(*dataset_ref);
  } else {
    /*
     * We didn't find it, so we need to look up the dataset_t on disk, and
     * create the dataset_ref_t.
     */
    memset(&key, 0, sizeof(key));
    key.data = &dataset_key;
    key.size = sizeof(dataset_key_t);

    memset(&value, 0, sizeof(value));
    value.flags = DB_DBT_MALLOC;

    result = db->db->get(db->db, trx_DB_TXN(trx), &key, &value, 0);
    if (result == OK) {
      VL(("\tFound it: Created dataset %p\n", value.data));
      result = dataset_ref_alloc(db,
				 hashentry,
				 dataset_key,
				 (dataset_t *) value.data,
				 dataset_ref);
      if (result == OK) {
	hash_SET_VALUE(hashentry, *dataset_ref);
      }
    } else {
      VL(("\tDidn't find it.\n"));
      if (result == DB_NOTFOUND)
	result = ACAP_ERR_NO_SUCH_DATASET;
      else if (result < 0)
	result = ACAP_ERR_DB;
    }
    if (result != OK)
      hash_delete_entry(hashentry);
  }

cleanup_lock:
  pthread_mutex_unlock(dataset_db_MUTEX(db));

  return result;
}

result_t
dataset_db_store(dataset_db_t * db,
		 trx_t * trx,
		 dataset_key_t * dataset_key,
		 dataset_t * dataset)
{
  result_t        result = OK;

  DBT             key,
                  value;

  I(db);
  I(db->db);
  I(dataset_key);

  VL(("Mapping dataset %p in dataset_db %p\n",
      dataset, db));

  memset(&key, 0, sizeof(key));
  key.data = dataset_key;
  key.size = sizeof(dataset_key_t);

  memset(&value, 0, sizeof(value));
  value.data = dataset;
  value.size = dataset ? dataset_SIZE(dataset) : 0;

  /*
   * Again, we wrap all this with a pthreads lock, because DB will lock us
   * anyway, and I'd rather make this fast if we're going to block at all.
   */
  pthread_mutex_lock(dataset_db_MUTEX(db));

  /*
   * Basically: I don't care *what* the key is, as long as it's *something*.
   */
  do {
    *dataset_key = mrand48();

    if (*dataset_key == 0)	/* 0 => magic; see entry_ref.c */
      continue;

    VL(("\tAttempting key %u...\n", *dataset_key));
    result = db->db->put(db->db, trx_DB_TXN(trx),
			 &key, &value, DB_NOOVERWRITE);
  } while (result == DB_KEYEXIST);

  pthread_mutex_unlock(dataset_db_MUTEX(db));

  VL(("\tUsed key %u\n", *dataset_key));

  if (result < 0)
    result = ACAP_ERR_DB;

  return result;
}

result_t
dataset_db_replace(dataset_db_t * db,
		   trx_t * trx,
		   dataset_key_t dataset_key,
		   dataset_t * dataset)
{
  result_t        result;

  DBT             key,
                  value;

  I(db);
  I(db->db);
  I(dataset_key);
  I(dataset);

  VL(("Mapping dataset %p to key %u in dataset_db %p\n",
      dataset, dataset_key, db));

  memset(&key, 0, sizeof(key));
  key.data = &dataset_key;
  key.size = sizeof(dataset_key_t);

  memset(&value, 0, sizeof(value));
  value.data = dataset;
  value.size = dataset_SIZE(dataset);

  pthread_mutex_lock(dataset_db_MUTEX(db));
  result = db->db->put(db->db, trx_DB_TXN(trx), &key, &value, 0);
  pthread_mutex_unlock(dataset_db_MUTEX(db));

  if (result < 0)
    result = ACAP_ERR_DB;

  return result;
}

result_t
dataset_db_remove(dataset_db_t * db,
		  trx_t * trx,
		  dataset_key_t dataset_key)
{
  result_t        result;

  DBT             key;

  I(db);
  I(db->db);

  VL(("Removing key %u in dataset_db %p\n",
      dataset_key, db));

  memset(&key, 0, sizeof(key));
  key.data = &dataset_key;
  key.size = sizeof(dataset_key_t);

  pthread_mutex_lock(dataset_db_MUTEX(db));
  result = db->db->del(db->db, trx_DB_TXN(trx), &key, 0);
  pthread_mutex_unlock(dataset_db_MUTEX(db));

  if (result < 0)
    result = ACAP_ERR_DB;


  return result;
}

result_t
dataset_db_check_norefs(dataset_db_t * db,
			dataset_ref_t * dataset_ref)
{
  I(db);
  I(dataset_ref);

  pthread_mutex_lock(dataset_db_MUTEX(db));
  refcount_UNREF_ELSE(dataset_ref_REFCOUNT(dataset_ref),
		      (hash_delete_entry(dataset_ref_HASHENT(dataset_ref)),
		       pthread_mutex_unlock(dataset_db_MUTEX(db)),
		       dataset_ref_free(&dataset_ref)),
		      pthread_mutex_unlock(dataset_db_MUTEX(db)));
  return OK;
}

result_t
dataset_db_sync(dataset_db_t * db)
{
  result_t        result;

  I(db);
  I(db->db);

  VL(("Synchronizing dataset_db %p\n", db));

  pthread_mutex_lock(dataset_db_MUTEX(db));
  result = db->db->sync(db->db, 0);
  pthread_mutex_unlock(dataset_db_MUTEX(db));

  if (result < 0)
    result = ACAP_ERR_DB;

  return result;
}

static result_t
dataset_db_alloc_key(dataset_key_t dataset_key,
		     const str_t * name,
		     void **dest,
		     uint32_t * dest_len)
{
  I(name);

  *dest_len = sizeof(dataset_key_t) + str_MEM_NEEDED(str_LEN(name));

  *dest = malloc(*dest_len);
  if (!*dest)
    return errno;

  memcpy(*dest, &dataset_key, sizeof(dataset_key));
  memcpy((char *) *dest + sizeof(dataset_key),
	 name,
	 str_MEM_NEEDED(str_LEN(name)));

  return OK;
}

result_t
dataset_db_lookup_entry(dataset_db_t * db,
			trx_t * trx,
			dataset_key_t dataset_key,
			const str_t * name,
			dataset_key_t * entry_key)
{
  result_t        result;

  DBT             key,
                  value;

  I(db);
  I(trx);
  I(dataset_key != UNDETERMINED_DATASET_KEY);
  I(name);

  memset(&key, 0, sizeof(key));
  result = dataset_db_alloc_key(dataset_key, name,
				&key.data, &key.size);
  if (result != OK)
    return result;

  memset(&value, 0, sizeof(value));
  value.data = entry_key;
  value.ulen = sizeof(*entry_key);
  value.flags = DB_DBT_USERMEM;

  pthread_mutex_lock(dataset_db_MUTEX(db));
  result = db->db->get(db->db, trx_DB_TXN(trx), &key, &value, 0);
  pthread_mutex_unlock(dataset_db_MUTEX(db));

  if (result == DB_NOTFOUND) {
    *entry_key = UNDETERMINED_DATASET_KEY;
    result = OK;
  }
  if (result < 0)
    result = ACAP_ERR_DB;

  free(key.data);

  return result;
}

result_t
dataset_db_replace_entry(dataset_db_t * db,
			 trx_t * trx,
			 dataset_key_t dataset_key,
			 const str_t * name,
			 dataset_key_t entry_key)
{
  result_t        result;

  DBT             key,
                  value;

  I(db);
  I(trx);
  I(dataset_key != UNDETERMINED_DATASET_KEY);
  I(name);
  I(entry_key != UNDETERMINED_DATASET_KEY);

  memset(&key, 0, sizeof(key));
  result = dataset_db_alloc_key(dataset_key, name,
				&key.data, &key.size);
  if (result != OK)
    return result;

  memset(&value, 0, sizeof(value));
  value.data = &entry_key;
  value.size = sizeof(entry_key);

  pthread_mutex_lock(dataset_db_MUTEX(db));
  result = db->db->put(db->db, trx_DB_TXN(trx), &key, &value, 0);
  pthread_mutex_unlock(dataset_db_MUTEX(db));

  if (result < 0)
    result = ACAP_ERR_DB;

  free(key.data);

  return result;
}

result_t
dataset_db_remove_entry(dataset_db_t * db,
			trx_t * trx,
			dataset_key_t dataset_key,
			const str_t * name)
{
  result_t        result;

  DBT             key;

  I(db);
  I(trx);
  I(dataset_key != UNDETERMINED_DATASET_KEY);
  I(name);

  memset(&key, 0, sizeof(key));
  result = dataset_db_alloc_key(dataset_key, name,
				&key.data, &key.size);
  if (result != OK)
    return result;

  pthread_mutex_lock(dataset_db_MUTEX(db));
  result = db->db->del(db->db, trx_DB_TXN(trx), &key, 0);
  pthread_mutex_unlock(dataset_db_MUTEX(db));

  if (result < 0)
    result = ACAP_ERR_DB;

  free(key.data);

  return result;
}
