/* dataset_ref.c--dataset_ref implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define DATASET_REF_C
#include <config.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include "dataset_ref.h"
#include "../util/util_err.h"
#include "../util/sllist.h"
#include "dataset_data.h"
#include "../util/refcache.h"
#include "../daemon/daemon.h"
#include "acap_err.h"
#include "dataset_db_fac.h"
#include "entry_ref.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_acapd_interlock)

static pthread_once_t dataset_ref_cache_init_once = PTHREAD_ONCE_INIT;

static refcache_t *dataset_ref_cache = NULL;

static void
dataset_ref_check_dataset_ref_free(void *rock)
{
  dataset_ref_t  *dataset_ref;

  dataset_ref = (dataset_ref_t *) rock;
  I(dataset_ref);

  dataset_db_check_norefs(dataset_ref_DB(dataset_ref),
			  dataset_ref);
}

static void
dataset_ref_cache_init(void)
{
  result_t result;

  result = refcache_alloc(DATASET_REF_DEFAULT_CACHE_SIZE,
			  &dataset_ref_check_dataset_ref_free,
			  &dataset_ref_cache);
  if (result != OK)
    daemon_fatal(result, "Initializing dataset reference cache");
}

result_t
dataset_ref_ref(dataset_ref_t * dataset_ref)
{
  I(dataset_ref);
  VL(("Adding a reference to dataset_ref %p\n", dataset_ref));
  return refcount_REF(&dataset_ref->refcount);
}

result_t
dataset_ref_free(dataset_ref_t ** dataset_ref)
{
  I(dataset_ref);
  I(*dataset_ref);

  I(!hash_N_ENTRIES(dataset_ref_ENTRIES(*dataset_ref)));

  VL(("Freeing dataset_ref %p\n", *dataset_ref));

  hash_destroy(dataset_ref_ENTRIES(*dataset_ref));
  pthread_mutex_destroy(dataset_ref_DATA_MUTEX(*dataset_ref));
  pthread_mutex_destroy(dataset_ref_ENTRIES_MUTEX(*dataset_ref));
  pthread_mutex_destroy(dataset_ref_DATASET_MUTEX(*dataset_ref));
  refcount_destroy(dataset_ref_REFCOUNT(*dataset_ref));
  if (dataset_ref_DB(*dataset_ref))
    dataset_db_unref(&dataset_ref_DB(*dataset_ref));
  if ((*dataset_ref)->dataset)
    dataset_free(&(*dataset_ref)->dataset);

  free(*dataset_ref);

  *dataset_ref = NULL;

  return OK;
}

result_t
dataset_ref_alloc(struct dataset_db_s * db,
		  hash_entry_t * hashent,
		  dataset_key_t key,
		  dataset_t * dataset,
		  dataset_ref_t ** dest)
{
  result_t        result;

  I(db);
  I(hashent);
  I(dest);
  I(dataset);
  I(key != UNDETERMINED_DATASET_KEY);
  I(dest);

  *dest = (dataset_ref_t *) malloc(sizeof(dataset_ref_t));
  if (!*dest) {
    result = errno;
    goto cleanup;
  }
  VL(("Allocating dataset_ref %p\n", *dest));

  (*dest)->hashent = hashent;
  (*dest)->data = NULL;

  result = refcount_init(dataset_ref_REFCOUNT(*dest));
  if (result)
    goto cleanup_data;

  result = pthread_mutex_init(dataset_ref_DATA_MUTEX(*dest), NULL);
  if (result)
    goto cleanup_refcount;

  result = pthread_mutex_init(dataset_ref_ENTRIES_MUTEX(*dest), NULL);
  if (result)
    goto cleanup_data_mutex;

  result = pthread_mutex_init(dataset_ref_DATASET_MUTEX(*dest), NULL);
  if (result)
    goto cleanup_entries_mutex;

  result = hash_init(&hash_str_procs,
		     dataset_ref_ENTRIES(*dest));
  if (result)
    goto cleanup_dataset_mutex;

  (*dest)->dataset = dataset;
  dataset = NULL;

  (*dest)->key = key;
  (*dest)->db = db;

  VL(("Using key %u, db %p, dataset %p for dataset_ref %p\n",
      key, db, (*dest)->dataset, *dest));

  if (dataset_ref_DB(*dest))
    dataset_db_ref(dataset_ref_DB(*dest));

  dataset_ref_ref(*dest);

  VL(("Allocated dataset_ref %p\n", *dest));

  return OK;

cleanup_dataset_mutex:
  pthread_mutex_destroy(dataset_ref_DATASET_MUTEX(*dest));

cleanup_entries_mutex:
  pthread_mutex_destroy(dataset_ref_ENTRIES_MUTEX(*dest));

cleanup_data_mutex:
  pthread_mutex_destroy(dataset_ref_DATA_MUTEX(*dest));

cleanup_refcount:
  refcount_destroy(dataset_ref_REFCOUNT(*dest));

cleanup_data:
  free(*dest);
  *dest = NULL;

cleanup:
  if (dataset)
    dataset_free(&dataset);

  return result;
}

result_t
dataset_ref_unref(dataset_ref_t ** dataset_ref)
{
  result_t        result = OK;

  I(dataset_ref);
  I(*dataset_ref);

  VL(("Decrementing refcount on dataset %p\n", *dataset_ref));

  if (dataset_ref_DB(*dataset_ref))
    result = refcount_UNREF_CHECKLAST(dataset_ref_REFCOUNT(*dataset_ref),
		       dataset_db_check_norefs(dataset_ref_DB(*dataset_ref),
					       *dataset_ref));
  else
    result = refcount_UNREF(dataset_ref_REFCOUNT(*dataset_ref),
			    dataset_ref_free(dataset_ref));


  *dataset_ref = NULL;		/* either way... */

  return result;
}

result_t
dataset_ref_set_cachesize(unsigned newsize)
{
  pthread_once(&dataset_ref_cache_init_once,
	       &dataset_ref_cache_init);
  return refcache_setcachesize(dataset_ref_cache, newsize);
}

result_t
dataset_ref_lookup_entry(dataset_ref_t * dataset_ref,
			 trx_t * trx,
			 const str_t * name,
			 entry_ref_t ** entry_ref)
{
  result_t        result = OK;

  hash_entry_t   *hash_entry;

  hash_did_insert_t inserted;

  I(dataset_ref);
  I(entry_ref);
  I(name);

  *entry_ref = NULL;

  VL(("Looking up entry_ref for \"%s\" in dataset_ref %p\n",
      str_DATA(name), dataset_ref));

  result = pthread_mutex_lock(dataset_ref_ENTRIES_MUTEX(dataset_ref));

  result = hash_find(&dataset_ref->entries,
		     name,
		     HASH_INSERT, &inserted,
		     &hash_entry);
  if (result != OK)
    goto cleanup_entries_mutex;

  if (inserted == HASH_DID_INSERT) {
    /* We need to allocate and lock the entry... */
    entry_ref_t    *new_entry_ref = NULL;

    dataset_key_t   entry_key;

    result = dataset_db_lookup_entry(dataset_ref_DB(dataset_ref),
				     trx,
				     dataset_ref_KEY(dataset_ref),
				     name,
				     &entry_key);
    if (result != OK)
      goto cleanup_hash;

    result = entry_ref_alloc(dataset_ref,
			     hash_entry,
			     NULL,
			     entry_key,
			     &new_entry_ref);

cleanup_hash:
    if (result == OK)
      hash_SET_VALUE(hash_entry, new_entry_ref);
    else
      hash_delete_entry(hash_entry);

    if (result != OK)
      goto cleanup_entries_mutex;

    *entry_ref = new_entry_ref;

  } else {
    I(hash_GET_VALUE(hash_entry));

    *entry_ref = (entry_ref_t *) hash_GET_VALUE(hash_entry);

    entry_ref_ref(*entry_ref);
  }

  VL(("Looked up \"%s\" in dataset_ref %p; got entry_ref %p\n",
      str_DATA(name), dataset_ref, *entry_ref));

  pthread_once(&dataset_ref_cache_init_once,
	       &dataset_ref_cache_init);

  refcache_used(dataset_ref_cache,
		dataset_ref_REFCOUNT(dataset_ref),
		dataset_ref);

cleanup_entries_mutex:
  pthread_mutex_unlock(dataset_ref_ENTRIES_MUTEX(dataset_ref));

  return result;
}

result_t
dataset_ref_check_entry_norefs(dataset_ref_t * dataset_ref,
			       entry_ref_t * entry_ref)
{
  I(dataset_ref);
  I(entry_ref);

  pthread_mutex_lock(dataset_ref_ENTRIES_MUTEX(dataset_ref));
  refcount_UNREF_ELSE(entry_ref_REFCOUNT(entry_ref),
		      (hash_delete_entry(entry_ref_HASHENT(entry_ref)),
	       pthread_mutex_unlock(dataset_ref_ENTRIES_MUTEX(dataset_ref)),
		       entry_ref_free(&entry_ref)),
	      pthread_mutex_unlock(dataset_ref_ENTRIES_MUTEX(dataset_ref)));
  return OK;
}

result_t
dataset_ref_rename_entry(dataset_ref_t * dataset_ref __attribute__((unused)),
			 str_t * oldname __attribute__((unused)),
			 str_t * newname __attribute__((unused)))
{
  return ACAP_ERR_UNIMPLEMENTED;/* TODO */
}

result_t
dataset_ref_get_dataset(dataset_ref_t * dataset_ref,
			trx_t * trx,
			dataset_t ** dest)
{
  I(dataset_ref);
  I(trx);
  I(dest);

  /* Someday, we need to actually do a history traverse here... XXX */
  pthread_mutex_lock(dataset_ref_DATASET_MUTEX(dataset_ref));
  *dest = dataset_ref->dataset;
  pthread_mutex_unlock(dataset_ref_DATASET_MUTEX(dataset_ref));
  return OK;
}

result_t
dataset_ref_set_inherit(dataset_ref_t * dataset_ref,
			trx_t * trx,
			bool_t default_inherit,
			str_t * path)
{
  result_t        result;

  dataset_t      *dataset;

  I(dataset_ref);
  I(trx);
  I(trx_IS_RW(trx));
  I(!(default_inherit && path));

  pthread_mutex_lock(dataset_ref_DATASET_MUTEX(dataset_ref));

  /*
   * Someday, we need to actually take care of history management somewhere
   * in here... XXX
   */

  result = dataset_rewrite(dataset_ref->dataset,
			   trx_STAMP(trx),
			   dataset_DATAPATH(dataset_ref->dataset),
			   default_inherit,
			   path,
			   dataset_BASE_ACL(dataset_ref->dataset),
			   &dataset);
  if (result != OK)
    goto cleanup_mutex;

  result = dataset_db_replace(dataset_ref->db,
			      trx,
			      dataset_ref->key,
			      dataset);
  if (result != OK)
    goto cleanup_dataset;

  {
    dataset_t      *temp;

    temp = dataset_ref->dataset;
    dataset_ref->dataset = dataset;
    dataset = temp;
  }

cleanup_dataset:
  if (dataset)
    dataset_free(&dataset);

cleanup_mutex:
  pthread_mutex_unlock(dataset_ref_DATASET_MUTEX(dataset_ref));

  return result;
}

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_acapd_dataset)

result_t
dataset_ref_get_dataset_data(dataset_ref_t * dataset_ref,
			     dataset_data_t ** dataset_data)
{
  result_t        result = OK;

  I(dataset_ref);
  I(dataset_data);

  VL(("Getting dataset data for dataset_ref %p\n", dataset_ref));

  pthread_mutex_lock(dataset_ref_DATA_MUTEX(dataset_ref));
  if (dataset_ref_DATA(dataset_ref)) {
    VL(("Using cached data object\n"));
    *dataset_data = dataset_ref_DATA(dataset_ref);
    dataset_data_ref(*dataset_data);
    pthread_mutex_unlock(dataset_ref_DATA_MUTEX(dataset_ref));
    return OK;
  }
  VL(("Using path \"%s\"\n",
      str_DATA(dataset_DATAPATH(dataset_ref->dataset))));
  result = dataset_data_alloc(str_DATA(dataset_DATAPATH(dataset_ref->dataset)),
			      dataset_ref, dataset_data);

  if (result == OK) {
    dataset_ref->data = *dataset_data;
    VL(("Created dataset_data %p\n", *dataset_data));
  }
  pthread_mutex_unlock(dataset_ref_DATA_MUTEX(dataset_ref));

  pthread_once(&dataset_ref_cache_init_once,
	       &dataset_ref_cache_init);

  refcache_used(dataset_ref_cache,
		dataset_ref_REFCOUNT(dataset_ref),
		dataset_ref);

  return result;
}

result_t
dataset_ref_check_data_norefs(dataset_ref_t * dataset_ref,
			      refcount_t * refcount,
			      dataset_data_t * dataset_data)
{
  I(dataset_ref);
  I(dataset_data);

  VL(("Checking whether we can free dataset_data %p\n", dataset_data));

  pthread_mutex_lock(dataset_ref_DATA_MUTEX(dataset_ref));
  I(dataset_ref_DATA(dataset_ref) == dataset_data);
  refcount_UNREF_ELSE(refcount,
		      (dataset_ref->data = NULL,
		  pthread_mutex_unlock(dataset_ref_DATA_MUTEX(dataset_ref)),
		       dataset_data_free(&dataset_data)),
		 pthread_mutex_unlock(dataset_ref_DATA_MUTEX(dataset_ref)));

  return OK;
}
