/* entry.h - entry interface
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#ifndef ACAP_ENTRY_H
#define ACAP_ENTRY_H

#include "stamp.h"
#include "../util/str.h"
#include "attr.h"

/* This is an ACAP entry.  It uses network byte order for everything
 * (the upshot: portability between all systems with a 32-bit integer
 * type), and may be moved arbitrarily around in memory - all internal
 * pointers are merely offsets. */

typedef struct {
  hashvalue_t     name_hash;	/* attribute name hash, NBO */
  uint32_t        offset;	/* offset from here, NBO */
}               attr_ref_t;

typedef struct entry_s {
  uint32_t        size;		/* amount of memory used by the entry (NBO) */
  stamp_t         modtime;
  uint32_t	  flags;	/* entry flags, NBO */
  dataset_key_t   subdataset;	/* subdataset key, NBO */
  uint32_t        n_attrs;	/* number of attributes, NBO */
  uint32_t        attr_refs_offset;
  /*
   * This is immediately followed, in memory, by the attributes (attr_t's).
   */
}               entry_t;

/* These flags are subtly different.  The idea is that when an entry
 * first has a NIL stored to it, ENTRY_FLAG_VISIBLE is clear and
 * ENTRY_FLAG_INHERIT is clear; this causes the entry to stop
 * inheriting, and causes it to not show up in searches.  As soon as
 * another store is performed, ENTRY_FLAG_VISIBLE is set, so the entry
 * will show up in search results (independant of ENTRY_FLAG_INHERIT).
 * A vanilla entry holding some data has ENTRY_FLAG_INHERIT set and
 * ENTRY_FLAG_VISIBLE set.  Note that no entry should *ever* have
 * ENTRY_FLAG_VISIBLE clear and ENTRY_FLAG_INHERIT set... */

#define ENTRY_FLAG_INHERIT (0x01)
#define ENTRY_FLAG_VISIBLE (0x02)

typedef struct {
  attr_ref_t     *loc;
  attr_ref_t     *end;
}               entry_iterator_t;

/* Accessor macros */
#define entry_SIZE(e) (ntohl((e)->size))
#define entry_ATTR_REFS(e) ((attr_ref_t *)((caddr_t)(&(e)->attr_refs_offset) \
					   + ntohl((e)->attr_refs_offset)))
#define entry_ITER_BEGIN(e, i) ((i)->loc = entry_ATTR_REFS(e)	\
				+ ntohl((e)->n_attrs) - 1,	\
				(i)->end = entry_ATTR_REFS(e) - 1)
#define entry_ITER_ATEND(i) ((i)->loc <= (i)->end)
#define entry_ITER_INCR(i) ((i)->loc--)
#define entry_ITER_VALUE(i) ((i)->loc)
#define entry_N_ATTRS(e) (ntohl((e)->n_attrs))
#define entry_ATTR(i) ((attr_t *)((caddr_t)(&(i)->offset)	\
				  + ntohl((i)->offset)))
#define entry_ATTR_NAME_HASH(i) ((hashvalue_t) ntohl((i)->name_hash))
#define entry_MODTIME(e) (&((e)->modtime))
#define entry_FLAGS(e) (ntohl((e)->flags))
#define entry_INHERITS(e) (entry_FLAGS(e) & ENTRY_FLAG_INHERIT)
#define entry_VISIBLE(e) (entry_FLAGS(e) & ENTRY_FLAG_VISIBLE)
#define entry_HAS_SUBDATASET(e) (ntohl((e)->subdataset) != UNDETERMINED_DATASET_KEY)
#define entry_SUBDATASET(e) (ntohl((e)->subdataset))

typedef struct {
  uint32_t        size;		/* amount of memory actually alloced */
}               entry_create_state_t;

/* Construction */
extern          result_t
entry_alloc(const stamp_t * modtime,
	    uint32_t flags,
	    dataset_key_t subdataset,
	    entry_t ** ent,
	    entry_create_state_t * create_state);

/* These are used for appending data to entries.  First, an attribute
 * is appended; ACLs, values, and a ctype are set (ACLs before
 * values).  A new attribute may be started by calling
 * entry_add_attribute; at the end, entry_complete() should be called.
 * The caller should allocate an entry_create_state_t on the stack,
 * and use it while building an entry. */

/* Note that the caller's responsible for appending the entry and
 * modtime attributes (in that order, before all others).  This is
 * intentional - it's a lot easier to trust the caller to do this than
 * to force the caller to provide all the relevant info to
 * entry_alloc(). */

extern result_t
entry_add_attribute(const str_t * name,
		    entry_t ** ent,
		    attr_state_t state,
		    entry_create_state_t * create_state);

extern result_t
entry_add_attr_acl(const str_t * acl,
		   entry_t ** ent,
		   entry_create_state_t * create_state);

extern result_t
entry_add_attr_value(const str_t * value,
		     entry_t ** ent,
		     entry_create_state_t * create_state);

extern result_t
entry_add_attr_ctype(const str_t * ctype,
		     entry_t ** ent,
		     entry_create_state_t * create_state);

extern result_t
entry_complete(entry_t ** ent,
	       entry_create_state_t * create_state);

/* This had *better* only be called for entries which're on the heap... */
extern result_t entry_free(entry_t ** ent);

/* Get an entry's name. */
extern result_t
entry_get_name(entry_t * ent,
	       str_t ** name);

#endif				/* ACAP_ENTRY_H */
