/* parse.acap.y - acap parser
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

%{
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <com_err.h>
#include "../util/str.h"
#include "client.h"
#include "command.h"
#include "search_ast.h"
#include "search.h"
#include "store.h"
#include "acap_err.h"
#include "token.acap.h"
#include "criteria.h"

#define YYPARSE_PARAM _client
#define YYLEX_PARAM _client
#define YYERROR_VERBOSE

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (acap_debug)

#define CHECK_RESULT()				\
  if (result) {					\
    report_error_code(CLIENT, result);		\
    YYERROR;					\
  }

%}

%pure_parser
%token_table

%union {
  str_t *str;
  unsigned long number;
  bool_t boolean;

  struct search_modifier_s *search_modifier;
  struct search_return_s *search_return;
  struct search_return_metadatum_s *search_return_metadatum;
  struct search_sort_s *search_sort;
  struct criterion_s *criterion;

  struct store_s *store;
  struct store_params_s *store_params;
  struct store_entry_s *store_entry;
  struct store_attribute_s *store_attribute;
  struct store_attribute_value_s *store_attribute_value;
  struct store_metadatum_s *store_metadatum;
}

%{
extern int yylex(YYSTYPE *lvalp, void *client);
static void yyerror(char *s);
static void unimplemented(client_t *client);
static void report_error_code(client_t *client, result_t code);
static result_t add_client_data(client_t *client, void *data);

#define CLIENT ((client_t *)(_client))
%}

%token ACL
%token ALL
%token AND
%token ATTRIBUTE
%token AUTHENTICATE
%token COMPARE
%token COMPARESTRICT
%token CONTENT_TYPE
%token CRLF
%token DEFAULT
%token DELETEACL
%token DELETEDSINCE
%token DEPTH
%token ENUMERATE
%token EQUAL
%token ERROR
%token FREECONTEXT
%token GETQUOTA
%token HARDLIMIT
%token LANG
%token LIMIT
%token LISTRIGHTS
%token LOGOUT
%token MAKECONTEXT
%token MYRIGHTS
%token NIL
%token NOCREATE
%token NOINHERIT
%token NOOP
%token NOT
%token NOTIFY
%token NUMBER
%token OR
%token PREFIX
%token RANGE
%token RETURN
%token SEARCH
%token SETACL
%token SIZE
%token SORT
%token STORE
%token STRING
%token SUBSTRING
%token TAG
%token UNCHANGEDSINCE
%token UPDATECONTEXT
%token VALUE

%type <str> TAG STRING optional_string value_nildef value_nil search_path store_path
%type <number> NUMBER
%type <boolean> maybe_enumerate maybe_notify
%type <search_modifier> search_modifier
%type <search_return> search_return_list search_return_list_maybe search_return
%type <search_return_metadatum> search_return_metadatum_list search_return_metadatum_list_maybe search_return_metadatum
%type <search_sort> search_sort_list search_sort_list_maybe search_sort
%type <str> comparator
%type <criterion> criterion
%type <store> store_entry_list
%type <store_params> store_params
%type <store_entry> store_entry attribute_store_list
%type <store_attribute> attribute_store
%type <store_attribute> attribute_store_value
%type <store_attribute> metadatum_write_list
%type <store_metadatum> metadatum_write
%type <store_metadatum> metadatum_value
%type <store_metadatum> metadatum_string_list_maybe
%type <store_metadatum> metadatum_string_list

%%

top: { CLIENT->recovering = FALSE; CLIENT->isbad = FALSE; } line
{
  return FALSE;
}

line: tag ' ' command
    | error CRLF
{
  client_parser_do_cleanup(CLIENT);
  CLIENT->recovering = FALSE;	/* no longer recovering */
}

tag: TAG
{
  client_SET_TAG(CLIENT, $1);
}

command: authenticate
       | deleteacl { unimplemented(CLIENT); }
       | deletedsince { unimplemented(CLIENT); }
       | freecontext
       | getquota { unimplemented(CLIENT); }
       | lang { unimplemented(CLIENT); }
       | listrights { unimplemented(CLIENT); }
       | logout
       | myrights { unimplemented(CLIENT); }
       | noop
       | search
       | setacl { unimplemented(CLIENT); }
       | store
       | updatecontext { unimplemented(CLIENT); }

authenticate: AUTHENTICATE check_no_auth ' ' STRING optional_string CRLF
{
  str_t *mechanism = $4;
  str_t *client_init_challenge = $5;

  YYSTYPE client_data, crlf_data;
  const char *clientin = NULL;
  unsigned clientinlen = 0;
  char *serverout;
  const char *errstr;
  unsigned serveroutlen;
  int sasl_result;
  bool_t do_yyerror = FALSE;

  I(mechanism);

  if (client_init_challenge) {
    clientin = str_DATA(client_init_challenge);
    clientinlen = str_LEN(client_init_challenge);
  }

  serverout = NULL;
  sasl_result = sasl_server_start(CLIENT->sasl, str_DATA(mechanism),
				  clientin, clientinlen,
				  &serverout, &serveroutlen,
				  &errstr);
  if (client_init_challenge)
    free(client_init_challenge);

  while (sasl_result == SASL_CONTINUE) {
    /* Send the data to the client, and recieve the next string. */
    pthread_mutex_lock(client_OUTPUT_MUTEX(CLIENT));
    if (! serveroutlen) {
      /* We don't have any data to send */
      client_printf(CLIENT, "+\r\n");
    } else {
      /* Send+free the data... */
      I(serverout);
      client_printf(CLIENT, "+ %l\r\n",
		    serverout, serveroutlen);
    }
    if (serverout) {
      free(serverout);
      serverout = NULL;
    }
    conn_flush(client_CONN(CLIENT));
    pthread_mutex_unlock(client_OUTPUT_MUTEX(CLIENT));

    /* And then wait for a response. */
    sasl_idle(CLIENT->sasl);

    client_SET_STATE(CLIENT, CLIENT_NORMAL);   /* *not* TAG */
    {
      int token = yylex(&client_data, CLIENT);
      if (token != '*' && token != STRING) {
	client_SET_BAD(CLIENT);
	report_error_code(CLIENT, ACAP_ERR_AUTH_NOT_STRING);
	do_yyerror = TRUE;
	sasl_result = SASL_FAIL;
	break;
      }

      if (yylex(&crlf_data, CLIENT) != CRLF) {
	client_SET_BAD(CLIENT);
	report_error_code(CLIENT, ACAP_ERR_AUTH_NOT_CRLF);
	do_yyerror = TRUE;
	sasl_result = SASL_FAIL;
	break;
      }
      if (token == '*') {
	report_error_code(CLIENT, ACAP_ERR_AUTH_CANCELED);
	do_yyerror = TRUE;
	sasl_result = SASL_FAIL;
	break;
      }
    }
    sasl_result = sasl_server_step(CLIENT->sasl,
				   str_DATA(client_data.str),
				   str_LEN(client_data.str),
				   &serverout, &serveroutlen,
				   &errstr);
    free(client_data.str);
  }

  if (sasl_result == SASL_OK)
    conn_SET_MECH(client_CONN(CLIENT), mechanism);
  else
    str_free(&mechanism);

  if (do_yyerror)
    YYERROR;

  I(client_TAG(CLIENT));

  pthread_mutex_lock(client_OUTPUT_MUTEX(CLIENT));
  if (sasl_result == SASL_OK) {
    /* We win! */
    const char *username;
    client_printf(CLIENT, "%t Ok", str_DATA(client_TAG(CLIENT)));
    if (serverout) {
      client_printf(CLIENT, " (SASL %l)", serverout, serveroutlen);
      free(serverout);
    }
    client_printf(CLIENT, " \"Authenticate Complete\"\r\n");
    conn_set_sasl(client_CONN(CLIENT), CLIENT->sasl);
    sasl_result = sasl_getprop(CLIENT->sasl, SASL_USERNAME,
			       (void **)&username);
    if (sasl_result == OK && username) {
      str_t *usernamestr = NULL;
      result_t result;
      result = str_copy_cstr(username, &usernamestr);
      if (result == OK)
	CLIENT->username = usernamestr;
    }
    CLIENT->authenticated = TRUE;
  } else if (sasl_result) {
    client_printf(CLIENT, "%t No %z\r\n",
		  str_DATA(client_TAG(CLIENT)),
		  errstr ? errstr : "Authentication Failed");
  }
  pthread_mutex_unlock(client_OUTPUT_MUTEX(CLIENT));

  free(client_TAG(CLIENT));

  client_parser_forget_cleanup(CLIENT);

  return FALSE;			/* don't close the connection */
}

deleteacl: DELETEACL check_auth ' ' acl_object optional_string CRLF

deletedsince: DELETEDSINCE check_auth ' ' STRING STRING CRLF

freecontext: FREECONTEXT check_auth ' ' STRING CRLF
{
  I(client_TAG(CLIENT));
  client_send_result(CLIENT, client_TAG(CLIENT),
		     client_forget_context(CLIENT, $4),     
		     "FreeContext Complete", NULL);
  client_parser_do_cleanup(CLIENT);
  return FALSE;			/* don't close the connection */
}

getquota: GETQUOTA check_auth ' ' STRING CRLF

lang: LANG ' ' lang_tag_list CRLF

lang_tag_list: lang_tag
	     | lang_tag_list ' ' lang_tag

lang_tag: STRING
{}

listrights: LISTRIGHTS check_auth ' ' acl_object ' ' STRING CRLF

logout: LOGOUT CRLF
{
  I(client_TAG(CLIENT));
  pthread_mutex_lock(client_OUTPUT_MUTEX(CLIENT));
  client_printf(CLIENT, "* Bye \"Cya Later!\"\r\n");
  client_printf(CLIENT, "%t Ok \"Logout Complete\"\r\n",
		str_DATA(client_TAG(CLIENT)));
  pthread_mutex_unlock(client_OUTPUT_MUTEX(CLIENT));
  client_parser_do_cleanup(CLIENT);
  return TRUE;			/* *do* close the connection */
}

myrights: MYRIGHTS check_auth ' ' acl_object CRLF

noop: NOOP CRLF
{
  I(client_TAG(CLIENT));
  client_send_result(CLIENT, client_TAG(CLIENT), OK, "NoOp Complete", NULL);
  client_parser_do_cleanup(CLIENT);
  return FALSE;			/* don't close the connection */
}

search: SEARCH check_auth ' ' search_path search_modifier ' ' criterion CRLF
{
  result_t result = search_start(CLIENT, $4, $5, $7);
  if (result) {
    pthread_mutex_lock(client_OUTPUT_MUTEX(CLIENT));
    client_printf(CLIENT, "%t No %z\r\n",
		  str_DATA(client_TAG(CLIENT)),
		  error_message(result));
    pthread_mutex_unlock(client_OUTPUT_MUTEX(CLIENT));
  } else {
    /* The search has taken all responsibility for this command's data */
    client_parser_forget_cleanup(CLIENT);
  }
  return FALSE;
}

search_path: STRING
{
  if (str_LEN($1) == 0) {
    client_SET_BAD(CLIENT);
    report_error_code(CLIENT, ACAP_ERR_BAD_PATH_TOO_SHORT);
    YYERROR;
  }
  $$ = $1;
}

search_modifier: /* NULL */
{
  result_t result = search_modifier_alloc(CLIENT, &$$);
  CHECK_RESULT();
}
	       | search_modifier ' ' DEPTH ' ' NUMBER
{
  result_t result = search_modifier_set_depth($1, CLIENT, $5);
  CHECK_RESULT();
  $$ = $1;
}
	       | search_modifier ' ' HARDLIMIT ' ' NUMBER
{
  result_t result = search_modifier_set_hardlimit($1, CLIENT, $5);
  CHECK_RESULT();
  $$ = $1;
}
	       | search_modifier ' ' LIMIT ' ' NUMBER ' ' NUMBER
{
  result_t result = search_modifier_set_limit($1, CLIENT, $5, $7);
  CHECK_RESULT();
  $$ = $1;
}
	       | search_modifier ' ' MAKECONTEXT ' ' maybe_enumerate maybe_notify STRING
{
  result_t result = search_modifier_set_makecontext($1, CLIENT, $5, $6, $7);
  CHECK_RESULT();
  $$ = $1;
}
	       | search_modifier ' ' NOINHERIT
{
  result_t result = search_modifier_set_noinherit($1, CLIENT);
  CHECK_RESULT();
  $$ = $1;
}
	       | search_modifier ' ' RETURN ' ' '(' search_return_list_maybe ')'
{
  result_t result = search_modifier_set_return($1, CLIENT, $6);
  CHECK_RESULT();
  $$ = $1;
}
	       | search_modifier ' ' SORT ' ' '(' search_sort_list_maybe ')'
{
  result_t result = search_modifier_set_sort($1, CLIENT, $6);
  CHECK_RESULT();
  $$ = $1;
}

search_return_list_maybe: /* NULL */ { $$ = NULL; }
			| search_return_list { $$ = $1; }

search_return_list: search_return
{
  $$ = $1;
}
		  | search_return_list ' ' search_return
{
  result_t result = search_return_list_append($1, $3, &$$);
  CHECK_RESULT();
}

search_return: STRING search_return_metadatum_list_maybe
{
  result_t result = search_return_alloc(CLIENT, $1, $2, &$$);
  CHECK_RESULT();
}

search_return_metadatum_list_maybe: /* NULL */ { $$ = NULL; }
			    | '(' search_return_metadatum_list ')' { $$ = $2; }

search_return_metadatum_list: search_return_metadatum
{
  $$ = $1;
}
			   | search_return_metadatum_list ' ' search_return_metadatum
{
  result_t result = search_return_metadatum_list_append($1, $3, &$$);
  CHECK_RESULT();
}

search_return_metadatum: STRING
{
  result_t result = search_return_metadatum_alloc(CLIENT, $1, &$$);
  CHECK_RESULT();
}

search_sort_list_maybe: /* NULL */ { $$ = NULL; }
		      | search_sort_list { $$ = $1; }

search_sort_list: search_sort
{
  $$ = $1;
}
		| search_sort_list ' ' search_sort
{
  result_t result = search_sort_list_append($1, $3, &$$);
  CHECK_RESULT();
}

search_sort: STRING ' ' STRING
{
  result_t result = search_sort_alloc(CLIENT, $1, $3, &$$);
  CHECK_RESULT();
}
		
maybe_enumerate: /* NULL */ { $$ = FALSE; }
	       | ENUMERATE ' ' { $$ = TRUE; }

maybe_notify: /* NULL */ { $$ = FALSE; }
	    | NOTIFY ' ' { $$ = TRUE; }

criterion: ALL
{
  result_t result = criterion_all_alloc(&$$);
  if (result == OK) result = add_client_data(CLIENT, $$);
  CHECK_RESULT();
}
	       | AND ' ' criterion ' ' criterion
{
  result_t result = criterion_and_alloc($3, $5, &$$);
  if (result == OK) result = add_client_data(CLIENT, $$);
  CHECK_RESULT();
}
	       | COMPARE ' ' STRING ' ' comparator ' ' value_nil
{
  result_t result = criterion_compare_alloc($3, $5, $7, &$$);
  if (result == OK) result = add_client_data(CLIENT, $$);
  CHECK_RESULT();
}
	       | COMPARESTRICT ' ' STRING ' ' comparator ' ' value_nil
{
  result_t result = criterion_comparestrict_alloc($3, $5, $7, &$$);
  if (result == OK) result = add_client_data(CLIENT, $$);
  CHECK_RESULT();
}
	       | EQUAL ' ' STRING ' ' comparator ' ' value_nil
{
  result_t result = criterion_equal_alloc($3, $5, $7, &$$);
  if (result == OK) result = add_client_data(CLIENT, $$);
  CHECK_RESULT();
}
	       | NOT ' ' criterion
{
  result_t result = criterion_not_alloc($3, &$$);
  if (result == OK) result = add_client_data(CLIENT, $$);
  CHECK_RESULT();
}
	       | OR ' ' criterion ' ' criterion
{
  result_t result = criterion_or_alloc($3, $5, &$$);
  if (result == OK) result = add_client_data(CLIENT, $$);
  CHECK_RESULT();
}
	       | PREFIX ' ' STRING ' ' comparator ' ' value_nil
{
  result_t result = criterion_prefix_alloc($3, $5, $7, &$$);
  if (result == OK) result = add_client_data(CLIENT, $$);
  CHECK_RESULT();
}
	       | RANGE ' ' NUMBER ' ' NUMBER ' ' STRING
{
  result_t result = criterion_range_alloc($3, $5, $7, &$$);
  if (result == OK) result = add_client_data(CLIENT, $$);
  CHECK_RESULT();
}
	       | SUBSTRING ' ' STRING ' ' comparator ' ' value_nil
{
  result_t result = criterion_substring_alloc($3, $5, $7, &$$);
  if (result == OK) result = add_client_data(CLIENT, $$);
  CHECK_RESULT();
}

comparator: STRING
{
  if (! str_is_UTF8($1)) {
    client_SET_BAD(CLIENT);
    report_error_code(CLIENT, ACAP_ERR_COMPARATOR_NOT_UTF8);
    YYERROR;
  }
  $$ = $1;
}

setacl: SETACL check_auth ' ' acl_object ' ' STRING ' ' STRING CRLF

store: STORE check_auth store_entry_list CRLF
{
  result_t result = store_start(CLIENT, $3);
  if (result != OK) {
    pthread_mutex_lock(client_OUTPUT_MUTEX(CLIENT));
    client_printf(CLIENT, "%t No %z\r\n",
		  str_DATA(client_TAG(CLIENT)),
		  error_message(result));
    pthread_mutex_unlock(client_OUTPUT_MUTEX(CLIENT));
  } else {
    /* The search takes all responsibility for this command's data */
    client_parser_forget_cleanup(CLIENT);
  }
  return FALSE;
}

store_entry_list: store_entry
{
  /* Initial store creation */
  result_t result = store_alloc(CLIENT, $1, &$$);
  CHECK_RESULT();
}
	     | store_entry_list store_entry
{
  /* Appending another store-entry */
  result_t result = store_append($1, CLIENT, $2);
  CHECK_RESULT();
  $$ = $1;
}

store_entry: ' ' '(' store_path ' ' store_params attribute_store_list ')'
{
  result_t result = OK;

  result = store_entry_set_params($6, $5);
  if (result == OK)
    result = store_entry_set_path($6, CLIENT, $3);
  CHECK_RESULT();
  $$ = $6;
}

store_path: STRING {
  if (str_LEN($1) <= 0
      || str_DATA($1)[0] != '/') {
    client_SET_BAD(CLIENT);
    report_error_code(CLIENT, ACAP_ERR_BAD_PATH_NO_SLASH);
    YYERROR;
  }
  if (str_DATA($1)[str_LEN($1) - 1] == '/') {
    client_SET_BAD(CLIENT);
    report_error_code(CLIENT, ACAP_ERR_BAD_PATH_END_SLASH);
    YYERROR;
  }
  $$ = $1;
}

store_params: /* NULL */
{
  result_t result = store_params_alloc(CLIENT, &$$);
  CHECK_RESULT();
}
	    | store_params NOCREATE ' '
{
  result_t result = store_params_set_nocreate($1);
  CHECK_RESULT();
  $$ = $1;
}
	    | store_params UNCHANGEDSINCE STRING ' '
{
  result_t result = store_params_set_unchangedsince($1, $3);
  CHECK_RESULT();
  $$ = $1;
}

attribute_store_list: attribute_store
{
  result_t result = store_entry_alloc(CLIENT, $1, &$$);
  CHECK_RESULT();
}
		    | attribute_store_list ' ' attribute_store
{
  result_t result = store_entry_append($1, CLIENT, $3);
  CHECK_RESULT();
  $$ = $1;
}

attribute_store: STRING ' ' attribute_store_value
{
  result_t result = store_attribute_set_name($3, CLIENT, $1);
  CHECK_RESULT();
  $$ = $3;
}

attribute_store_value: value_nildef
{
  result_t result = store_attribute_value_alloc(CLIENT, $1, &$$);
  CHECK_RESULT();
}
		     | '(' metadatum_write_list ')'
{
  $$ = $2;			/* just pop it up... */
}

value_nildef: DEFAULT { $$ = (str_t *)-1; }
	    | value_nil { $$ = $1; }

value_nil: NIL { $$ = NULL; }
	 | STRING { $$ = $1; }

metadatum_write_list: metadatum_write
{
  /* Initial recursion */
  result_t result = store_attribute_alloc(CLIENT, $1, &$$);
  CHECK_RESULT();
}
		   | metadatum_write_list ' ' metadatum_write
{
  /* Adding a metadatum */
  result_t result = store_attribute_append($1, CLIENT, $3);
  CHECK_RESULT();
  $$ = $1;
}

metadatum_write: STRING ' ' metadatum_value
{
  result_t result;
  const acap_token_t *token;

  token = acap_token_lookup(str_DATA($1), str_LEN($1));
  if (! token) {
    client_SET_BAD(CLIENT);
    report_error_code(CLIENT, ACAP_ERR_BAD_METADATUM_TYPE);
    YYERROR;
  }
  result = store_metadatum_set_name($3, CLIENT, token->value);
  CHECK_RESULT();
  $$ = $3;
}

metadatum_value: STRING
{
  result_t result = store_metadatum_str_alloc(CLIENT, $1, &$$);
  CHECK_RESULT();
}
	      | NIL
{
  result_t result = store_metadatum_nil_alloc(CLIENT, &$$);
  CHECK_RESULT();
}
	      | DEFAULT
{
  result_t result = store_metadatum_default_alloc(CLIENT, &$$);
  CHECK_RESULT();
}
	      | '(' metadatum_string_list_maybe ')'
{
  $$ = $2;
}

metadatum_string_list_maybe: /* empty */ {
  result_t result = store_metadatum_multi_alloc(CLIENT, NULL, &$$);
  CHECK_RESULT();
}
			  | metadatum_string_list { $$ = $1; }

metadatum_string_list: STRING
{
  result_t result = store_metadatum_multi_alloc(CLIENT, $1, &$$);
  CHECK_RESULT();
}
		    | metadatum_string_list ' ' STRING
{
  result_t result = store_metadatum_multi_append($1, CLIENT, $3);
  CHECK_RESULT();
  $$ = $1;
}

updatecontext: UPDATECONTEXT check_auth ' ' context_list CRLF

context_list : context
	     | context_list ' ' context

context: STRING
{}


acl_object: '(' STRING ')'
	  | '(' STRING ' ' STRING ')'
	  | '(' STRING ' ' STRING ' ' STRING ')'

optional_string: /* NULL */ { $$=NULL; } | ' ' STRING { $$ = $2; }

check_no_auth: /* NULL */
{
  /* We just use this to check, inline, whether we're not authentic */
  if (CLIENT->authenticated) {
    pthread_mutex_lock(client_OUTPUT_MUTEX(CLIENT));
    client_printf(CLIENT, "%t Bad \"Already Authenticated\"\r\n",
		  str_DATA(client_TAG(CLIENT)));
    pthread_mutex_unlock(client_OUTPUT_MUTEX(CLIENT));
    YYERROR;
  }
}

check_auth: /* NULL */
{
  /* We just use this to check, inline, whether we're authentic */
  if (! CLIENT->authenticated) {
    pthread_mutex_lock(client_OUTPUT_MUTEX(CLIENT));
    client_printf(CLIENT, "%t Bad \"Authenticate First\"\r\n",
		  str_DATA(client_TAG(CLIENT)));
    pthread_mutex_unlock(client_OUTPUT_MUTEX(CLIENT));
    YYERROR;
  }
}

%%

static void
unimplemented(client_t *client)
{
  pthread_mutex_lock(client_OUTPUT_MUTEX(client));
  client_printf(client, "%t No \"Unimplemented\"\r\n",
		str_DATA(client_TAG(client)));
  conn_flush(client_CONN(client));
  pthread_mutex_unlock(client_OUTPUT_MUTEX(client));
  client_parser_do_cleanup(client);
}

static void
report_error(client_t *client, const char *msg)
{
  const char *tag;

  I(client);
  I(msg);

  if (client_TAG(client))
    tag = str_DATA(client_TAG(client));
  else
    tag = "*";

  pthread_mutex_lock(client_OUTPUT_MUTEX(client));
  client_printf(client, "%t %t %z\r\n",
		tag,
		client_IS_BAD(client) ? "Bad" : "No",
		msg);
  pthread_mutex_unlock(client_OUTPUT_MUTEX(client));

  client->recovering = TRUE;
}

static void
report_error_code(client_t *client, result_t code)
{
  report_error(client, error_message(code));
}

static void
yyerror(char *s)
{
  client_t *client = NULL;

  if (client_current(&client))
    return;

  client_SET_BAD(client);

  report_error(client, s);
}

static result_t
add_client_data(client_t *client,
		void *data)
{
  result_t result = client_parser_add_cleanup(client, &free, data);
  if (result != OK)
    free(data);
  return result;
}
