/* store.c - store command implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define STORE_C
#include <config.h>
#include <stdlib.h>
#include <errno.h>
#include <com_err.h>
#include "acap_err.h"
#include "client.h"
#include "store.h"
#include "parse.acap.h"
#include "command.h"
#include "traverse.h"
#include "dataset_data.h"
#include "trx.h"
#include "../daemon/daemon_err.h"
#include "../daemon/daemon.h"
#include "entry_ref.h"
#include "dataset_ref.h"
#include "dataset_data.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_acapd_store)

static const str_t *entry_str = (str_t *) "\0\0\0\005entry\0\0\0\0\0\0";

static          result_t
append_attribute(const store_attribute_t * store_attr,
		 attr_t * attr,
		 bool_t inherit,
		 entry_t ** new_entry,
		 entry_create_state_t * create_state)
{
  result_t        result;

  store_metadatum_t *meta;

  str_t          *name;

  attr_state_t    attr_state;

  I(store_attr || attr);
  I(new_entry);
  I(*new_entry);
  I(create_state);

  /*
   * This isn't exactly ugly code, but... it's not exactly clean, either. It
   * might be nice to make this more abstract or something someday. Maybe
   * some structure that'd contain the metadatum, a pointer to the function
   * to use to append it, and a function for getting the data out of the
   * attr_t?  Hrm.  XXX
   */

  VL(("Store: appending attribute\n"));

  if (store_attr) {
    meta = store_attribute_METADATUM(store_attr, METADATUM_VALUE);
    if (meta)
      attr_state = store_metadatum_STATE(meta);
    else
      attr_state = attr_STATE(attr);
  } else
    attr_state = attr_STATE(attr);

  VL(("Store: attribute state is %d\n", attr_state));

  if (attr_state == ATTR_NIL && !inherit) {
    VL(("Nil stored to a non-inheriting entry--not appending attr\n"));
    return OK;			/* skip NIL attrs if we're not inheriting */
  }
  if (attr_state == ATTR_DEFAULT && inherit) {
    VL(("Default stored to a non-inheriting entry--not appending attr\n"));
    return OK;			/* skip DEFAULT attrs if we're inheriting */
  }
  if (store_attr)
    name = store_attribute_NAME(store_attr);
  else
    name = attr_NAME(attr);

  VL(("Store: adding attribute \"%s\"\n", str_DATA(name)));

  result = entry_add_attribute(name, new_entry, attr_state, create_state);
  if (result)
    return result;

  if (attr_state == ATTR_DEFAULT || attr_state == ATTR_NIL)
    return OK;			/* no data on NIL or DEFAULT attrs */

  VL(("Store: adding attribute data\n"));

  meta = store_attr ? store_attribute_METADATUM(store_attr, METADATUM_ACL)
    : NULL;
  if (meta) {
    /* Get the ACL from the store */
    store_metadatum_iterator_t it;

    VL(("Store: using ACL from store\n"));

    for (store_metadatum_ITER_BEGIN(meta, &it);
	 !store_metadatum_ITER_ATEND(&it);
	 store_metadatum_ITER_INCR(&it)) {
      result = entry_add_attr_acl(store_metadatum_ITER_VALUE(&it),
				  new_entry,
				  create_state);
      if (result != OK)
	return result;
    }
  } else if (attr) {
    /* Use the ACL from the original */
    strlist_iterator_t it;

    VL(("Store: using ACL from original entry\n"));

    for (strlist_ITER_BEGIN(attr_ACL(attr), &it);
	 !strlist_ITER_ATEND(&it);
	 strlist_ITER_INCR(&it)) {
      result = entry_add_attr_acl(strlist_ITER_VALUE(&it),
				  new_entry,
				  create_state);
      if (result != OK)
	return result;
    }
  }
  VL(("Store: done with ACL\n"));

  VL(("Store: doing Value\n"));
  meta = store_attr ? store_attribute_METADATUM(store_attr, METADATUM_VALUE)
    : NULL;
  if (meta) {
    store_metadatum_iterator_t it;

    VL(("Store: Getting value from store\n"));
    for (store_metadatum_ITER_BEGIN(meta, &it);
	 !store_metadatum_ITER_ATEND(&it);
	 store_metadatum_ITER_INCR(&it)) {
      result = entry_add_attr_value(store_metadatum_ITER_VALUE(&it),
				    new_entry,
				    create_state);
      if (result != OK)
	return result;
    }
  } else if (attr) {
    /* Get the value from the original */
    strlist_iterator_t it;

    VL(("Store: Getting value from original\n"));
    for (strlist_ITER_BEGIN(attr_VALUE(attr), &it);
	 !strlist_ITER_ATEND(&it);
	 strlist_ITER_INCR(&it)) {
      result = entry_add_attr_value(strlist_ITER_VALUE(&it),
				    new_entry,
				    create_state);
      if (result != OK)
	return result;
    }
  }
  meta = store_attr ? store_attribute_METADATUM(store_attr,
						METADATUM_CONTENT_TYPE)
    : NULL;
  if (meta) {
    result = entry_add_attr_value(store_metadatum_VALUE(meta),
				  new_entry,
				  create_state);
    if (result != OK)
      return result;
  } else if (attr && attr_HAS_CTYPE(attr)) {
    /* Get the content-type from the original */
    result = entry_add_attr_acl(attr_CTYPE(attr),
				new_entry,
				create_state);
    if (result != OK)
      return result;
  }
  VL(("Store: Done appending attribute\n"));

  return OK;
}

static result_t
merge_entry(trx_t * trx, store_entry_t * store_entry,
	    entry_t * original, entry_t ** dest, str_t ** dest_name)
{
  result_t        result = OK;

  entry_t        *new_entry = NULL;

  store_entry_iterator_t s_it;

  entry_iterator_t e_it;

  entry_create_state_t create_state;

  store_attribute_t *store_attr;

  store_metadatum_t *store_metadatum;

  I(trx);
  I(store_entry);
  I(original);
  I(dest);
  I(!*dest);
  I(dest_name);

  VL(("Beginning entry merge\n"));

  /*
   * Okay, three possibilities: Storing a NIL to the entry attribute, storing
   * a DEFAULT to the entry attribute, neither of the above.
   */

  store_attr = store_entry_ENTRY_ATTR(store_entry);
  if (store_attr) {
    store_metadatum = store_attribute_METADATUM(store_attr, METADATUM_VALUE);
    if (store_metadatum) {
      if (store_metadatum_STATE(store_metadatum) == ATTR_NIL) {
	/* Storing a NIL to the entry attribute */
	VL(("Storing a NIL to the entry attribute\n"));
	/* TODO: Blow away the subdataset, if we had one... */
	result = entry_alloc(trx_STAMP(trx),
			     0,
			     UNDETERMINED_DATASET_KEY,
			     &new_entry,
			     &create_state);
	if (result != OK)
	  goto cleanup;
	result = entry_add_attribute(entry_str,
				     &new_entry,
				     ATTR_SINGLE,
				     &create_state);
	if (result != OK)
	  goto cleanup;
	result = entry_add_attr_value(store_entry_NAME(store_entry),
				      &new_entry,
				      &create_state);
	if (result != OK)
	  goto cleanup;
	result = entry_complete(&new_entry, &create_state);
	if (result != OK)
	  goto cleanup;
	*dest = new_entry;
	return OK;
      } else if (store_metadatum_STATE(store_metadatum) == ATTR_DEFAULT) {
	/* Storing a DEFAULT to the entry attribute */
	VL(("Storing a DEFAULT--removing the entry entirely\n"));
	*dest = NULL;
	*dest_name = NULL;
	return OK;
      } else {
	/* Storing a single value -- a rename. */
	VL(("Renaming the entry.\n"));
	I(store_metadatum_STATE(store_metadatum) != ATTR_MULTI);
	*dest_name = store_metadatum_VALUE(store_metadatum);
      }
    }
  }
  VL(("Constructing the entry...\n"));
  result = entry_alloc(trx_STAMP(trx),
		       entry_FLAGS(original) | ENTRY_FLAG_VISIBLE,
		       entry_SUBDATASET(original),
		       &new_entry,
		       &create_state);
  if (result != OK)
    goto cleanup;

  store_entry_ITER_BEGIN(store_entry, &s_it);
  entry_ITER_BEGIN(original, &e_it);
  VL(("Beginning attribute processing loop\n"));
  while (!store_entry_ITER_ATEND(&s_it)
	 || !entry_ITER_ATEND(&e_it)) {
    compare_result_t compare_result;

    VL(("Top of entry attribute loop\n"));

    if (store_entry_ITER_ATEND(&s_it)) {
      /* store_entry is finished; append from original */
      VL(("Store entry is finished; appending from original\n"));
      compare_result = COMPARE_LESS;
    } else if (entry_ITER_ATEND(&e_it)) {
      /* original is finished; append from store_entry */
      VL(("Original entry is finished; appending from store entry\n"));
      compare_result = COMPARE_GREATER;
    } else {
      /* neither is finished; decide what we're doing... */
      VL(("Comparing attrs: original is \"%s\", store is \"%s\"\n",
	  str_DATA(attr_NAME(entry_ATTR(entry_ITER_VALUE(&e_it)))),
	  str_DATA(store_attribute_NAME(store_entry_ITER_VALUE(&s_it)))));
      compare_result = str_compare(attr_NAME(entry_ATTR(entry_ITER_VALUE(&e_it))),
		       store_attribute_NAME(store_entry_ITER_VALUE(&s_it)));
      VL(("Compare: result was %d\n", compare_result));
    }
    /* Now that we have a decision... */
    switch (compare_result) {
    case COMPARE_LESS:
      /* Use the attribute from the original */
      VL(("Appending attribute from original\n"));
      result = append_attribute(NULL,
				entry_ATTR(entry_ITER_VALUE(&e_it)),
				entry_INHERITS(original),
				&new_entry,
				&create_state);
      if (result != OK)
	goto cleanup;
      VL(("Incrementing entry iterator\n"));
      entry_ITER_INCR(&e_it);
      break;
    case COMPARE_EQUAL:
      /* We're modifying an existing attribute. */
      VL(("Modifying existing attribute\n"));
      result = append_attribute(store_entry_ITER_VALUE(&s_it),
				entry_ATTR(entry_ITER_VALUE(&e_it)),
				entry_INHERITS(original),
				&new_entry,
				&create_state);
      VL(("Incrementing both iterators\n"));
      store_entry_ITER_INCR(&s_it);
      entry_ITER_INCR(&e_it);
      break;
    case COMPARE_GREATER:
      /* Use the attribute from the store */
      VL(("Appending attribute from store\n"));
      result = append_attribute(store_entry_ITER_VALUE(&s_it),
				NULL,
				entry_INHERITS(original),
				&new_entry,
				&create_state);
      if (result != OK)
	goto cleanup;
      VL(("Incrementing store entry iterator\n"));
      store_entry_ITER_INCR(&s_it);
      break;
    default:
      result = DAEMON_ERR_INTERNAL;
      goto cleanup;
    }
    VL(("Bottom of entry attribute loop\n"));
  }
  VL(("End of attribute processing loop\n"));

  result = entry_complete(&new_entry, &create_state);
  if (result != OK)
    goto cleanup;

  *dest = new_entry;

  return OK;

cleanup:
  if (new_entry)
    free(new_entry);

  return result;
}

static result_t
do_one_entry(client_t * client,
	     trx_t * trx,
	     str_t * modtime,
	     store_entry_t * store_entry,
	     respcode_t ** respcode)
{
  result_t        result = OK;

  entry_ref_t    *entry_ref = NULL;

  dataset_ref_t  *dataset_ref = NULL;

  dataset_data_t *dataset_data = NULL;

  entry_t        *original_entry = NULL,
                 *new_entry = NULL;

  str_t          *new_name = NULL;

  VL(("Traversing entry \"%s\"\n", str_DATA(store_entry_PATH(store_entry))));

  result = traverse_write(store_entry_PATH(store_entry),
			  client,
			  trx,
			  store_entry_CREATE(store_entry)
			  ? TRAVERSE_CREATE_DATASET
			  : TRAVERSE_NO_CREATE_DATASET,
			  &entry_ref,
			  respcode);

  if (result != OK)
    return result;

  /* At this point, we can actually set the store_entry's name... */
  result = store_entry_set_name(store_entry,
				entry_ref_NAME(entry_ref));
  if (result != OK)
    goto cleanup_entry_ref;

  /* And so we can complete the entry. */
  result = store_entry_complete(store_entry, modtime);
  if (result != OK)
    goto cleanup_entry_ref;

  VL(("Getting dataset_ref of \"%s\"\n", str_DATA(store_entry_PATH(store_entry))));

  result = entry_ref_get_parent_dataset(entry_ref, &dataset_ref);
  if (result != OK)
    goto cleanup_entry_ref;

  VL(("Getting dataset_data of \"%s\"\n", str_DATA(store_entry_PATH(store_entry))));

  result = dataset_ref_get_dataset_data(dataset_ref,
					&dataset_data);

  VL(("Got dataset_data %p\n", dataset_data));

  if (result != OK)
    goto cleanup_dataset_ref;

  VL(("Processing \"%s\" (\"%s\")\n",
      str_DATA(store_entry_PATH(store_entry)),
      str_DATA(store_entry_NAME(store_entry))));

  result = dataset_data_get_entry(dataset_data,
				  trx,
				  store_entry_NAME(store_entry),
				  &original_entry);

  if (result != OK)
    goto cleanup_dataset_data;

  new_name = store_entry_NAME(store_entry);

  result = merge_entry(trx, store_entry, original_entry, &new_entry, &new_name);

  if (result != OK)
    goto cleanup_new_entry;

  result = dataset_data_replace_entry(dataset_data,
				      trx,
				      store_entry_NAME(store_entry),
				      original_entry,
				      new_name,
				      new_entry);

  /*
   * The dataset_data is responsible for cleaning up the new entry at this
   * point.
   */

  new_entry = NULL;

  if (result != OK)
    goto cleanup_new_entry;

  if (str_equal(store_entry_NAME(store_entry), new_name) == EQ_NOT_EQUAL) {
    result = dataset_ref_rename_entry(dataset_ref,
				      store_entry_NAME(store_entry),
				      new_name);
    if (result != OK)
      goto cleanup_new_entry;
  }
  if (store_entry_INHERIT_VALUE(store_entry)) {
    store_metadatum_t *inherit;

    dataset_ref_t  *subds;

    respcode_t *respcode = NULL;

    result = traverse_get_dataset(entry_ref, trx, TRAVERSE_NO_CREATE_DATASET,
				  &subds, &respcode);
    if (respcode)		/* XXX Maybe we should report this? */
      respcode_free(&respcode);

    if (result == ACAP_ERR_NO_SUCH_DATASET) {
      result = OK;		/* but skip subds part... */
    } else {
      if (result != OK)
	goto cleanup_new_entry;

      inherit = store_entry_INHERIT_VALUE(store_entry);

      switch (store_metadatum_STATE(inherit)) {
      case ATTR_NIL:
	result = dataset_ref_set_inherit(subds, trx, FALSE, NULL);
	break;
      case ATTR_DEFAULT:
	result = dataset_ref_set_inherit(subds, trx, TRUE, NULL);
	break;
      case ATTR_SINGLE:
	result = dataset_ref_set_inherit(subds, trx, FALSE,
					 store_metadatum_VALUE(inherit));
	break;
      case ATTR_MULTI:
	daemon_fatal(DAEMON_ERR_INTERNAL,
		     "Multivalued dataset.inherit attribute");
	break;
      }
      dataset_ref_unref(&subds);
      if (result != OK)
	goto cleanup_new_entry;
    }
  }
  /*
   * Do we really need this? dataset_data_sync(dataset_data);
   */

cleanup_new_entry:
  if (new_entry)
    free(new_entry);

cleanup_dataset_data:
  if (original_entry)
    dataset_data_done_with_entry(dataset_data, &original_entry);

  dataset_data_unref(&dataset_data);

cleanup_dataset_ref:
  dataset_ref_unref(&dataset_ref);

cleanup_entry_ref:
  entry_ref_unref(&entry_ref);

  return result;
}

static void
store_run(void *_store)
{
  result_t        result = OK;

  store_t        *store = (store_t *) _store;

  store_iterator_t entry_iter;

  trx_t          *trx = NULL;

  str_t          *modtime = NULL;

  respcode_t     *respcode = NULL;

  VL(("Running store %p\n", store));

  result = trx_begin(&trx, TRX_RW);
  if (result != OK)
    goto cleanup;

  result = trx_str(trx, &modtime);
  if (result != OK)
    goto cleanup;

  if (result == OK)
    for (store_ITER_BEGIN(store, &entry_iter);
	 !store_ITER_ATEND(&entry_iter);
	 store_ITER_INCR(&entry_iter)) {
      result = do_one_entry(store->client, trx,
			    modtime, store_ITER_VALUE(&entry_iter),
			    &respcode);
      if (result)
	break;
    }

cleanup:
  if (modtime)
    str_free(&modtime);

  if (trx) {
    trx_drop_lock(trx);
    trx_done(&trx,
	     result == OK ? TRX_COMMIT : TRX_ABORT);
  }
  if (result == OK)
    VL(("Store %p complete\n", store));
  else
    VL(("Store %p returned %s\n", store, error_message(result)));

  client_send_result(store_CLIENT(store),
		     store_TAG(store),
		     result,
		     "Store Complete",
		     respcode);

  client_unref(&store_CLIENT(store));

  store_free(&store);
}

result_t
store_start(client_t * client,
	    store_t * store)
{
  result_t        result;

  dispatch_queue_t *queue = NULL;

  I(client);
  I(store);

  VL(("Queueing store %s\n", str_DATA(store_TAG(store))));

  result = command_get_writer_queue(&queue);
  if (result)
    return result;

  I(client == store_CLIENT(store));

  client_ref(client);

  result = dispatch_enqueue(queue, &store_run, store);
  if (result) {
    store_free(&store);
    client_unref(&client);
    return result;
  }
  VL(("Queued store %p\n", store));

  return OK;
}
