/* traverse.c--traverses ACAP paths
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define TRAVERSE_C
#include <config.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include "traverse.h"
#include "../util/rselock.h"
#include "../util/hash.h"
#include "../util/util_err.h"
#include "acap_err.h"
#include "dataset_ref.h"
#include "entry_ref.h"
#include "dataset_db.h"
#include "dataset_db_fac.h"
#include "dataset_data.h"
#include "entry.h"
#include "entry_ref_history.h"
#include "attr.h"
#include "respcode.h"
#include "../daemon/daemon.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_acapd_interlock)

static entry_ref_t *root_entry = NULL;

static pthread_once_t root_entry_init_once = PTHREAD_ONCE_INIT;

/* I really must write a macro to construct these someday... :-( */
static const str_t *entry_str = (str_t *) "\0\0\0\005entry\0\0\0\0\0\0";

static const str_t *modtime_str = (str_t *) "\0\0\0\007modtime\0\0\0\0";

static const str_t *subdataset_str = (str_t *) "\0\0\0\012subdataset\0";

static const str_t *subdataset_dot = (str_t *) "\0\0\0\001.\0\0";

static const str_t *user_str = (str_t *) "\0\0\0\004user\0\0\0\0\0\0\0";

static const str_t *slash_str = (str_t *) "\0\0\0\001/\0\0";

static const str_t *root_db_name_str = (str_t *) "\0\0\0\006master\0\0\0\0\0";

#define ROOT_DATASET_KEY (1)

static void
root_entry_init(void)
{
  result_t result;

  VL(("Initializing root entry_ref\n"));
  result = entry_ref_alloc(NULL, NULL, root_db_name_str,
			   ROOT_DATASET_KEY, &root_entry);
  if (result != OK)
    goto fatal;

  result = entry_ref_ref(root_entry);	/* this is never freed */
  if (result != OK)
    goto fatal;

fatal:
  if (result != OK)
    daemon_fatal(result, "Initializing dataset root entry information");
}

static result_t
get_root_entry_ref(entry_ref_t ** entry_ref)
{
  I(entry_ref);
  I(!*entry_ref);

  VL(("Looking up root entry_ref\n"));

  pthread_once(&root_entry_init_once,
	       &root_entry_init);

  *entry_ref = root_entry;

  entry_ref_ref(*entry_ref);

  return OK;
}

/* This defines an iterator to use when walking over paths;
 * this makes the path walking code much simpler... */
typedef struct {
  const char     *start,
                 *here,
                 *end;
}               traverse_split_iterator_t;

static result_t
traverse_set_noexist(const str_t ** begin,
		     const str_t ** end,
		     respcode_t ** dest)
{
  result_t        result;

  respcode_t     *code;

  str_t          *path;

  result = str_join(slash_str, slash_str, NULL, begin, end, &path);
  if (result != OK)
    return result;;

  result = respcode_noexist_alloc(path, &code);
  if (result != OK) {
    str_free(&path);
    return result;
  }
  if (*dest)
    respcode_free(dest);

  *dest = code;
  return OK;
}


static void
traverse_split_iterator_incr(traverse_split_iterator_t * it)
{
  I(it);

  it->start = ++(it->here);
  while (it->here < it->end && *(it->here) != '/')
    ++(it->here);
}

static void
traverse_split_iterator_init(const str_t * path,
			     traverse_split_iterator_t * it)
{
  I(path);
  I(it);

  it->here = str_DATA(path) - 1;/* init to just before data, so incr will
				 * make it point at data */
  it->end = str_DATA(path) + str_LEN(path);
  traverse_split_iterator_incr(it);
}

#define traverse_split_iterator_ATEND(i) ((i)->end <= (i)->start)
#define traverse_split_iterator_LEN(i) ((i)->here - (i)->start)
#define traverse_split_iterator_DATA(i) ((i)->start)

/* This splits a path into its components, for easier handling.
 * All of the memory is allocated in one blob. */
result_t
traverse_split_path(const str_t * path,
		    const str_t * username,	/* for ~ expansion */
		    unsigned append_len,
		    const str_t ** append,
		    unsigned *len,
		    const str_t *** elements)
{
  size_t          mem_needed;

  const str_t   **element,
                **append_it,
                **append_end;

  char           *data;

  unsigned        element_number;

  traverse_split_iterator_t it;

  I(path);
  I(len);
  I(elements);

  /* count... */
  mem_needed = 0;
  *len = 0;
  *elements = NULL;
  for (traverse_split_iterator_init(path, &it), element_number = 0;
       !traverse_split_iterator_ATEND(&it);
       traverse_split_iterator_incr(&it), ++element_number) {
    if (element_number == 2
	&& traverse_split_iterator_LEN(&it) == 1
	&& traverse_split_iterator_DATA(&it)[0] == '~') {
      *len += 2;
      mem_needed += 2 * sizeof(str_t *);
    } else {
      ++*len;
      mem_needed += str_MEM_NEEDED(traverse_split_iterator_LEN(&it));
      mem_needed += sizeof(str_t *);
    }
  }

  for (append_it = append, append_end = append + append_len;
       append_it < append_end;
       append_it++) {
    ++*len;
    mem_needed += str_MEM_NEEDED(str_LEN(*append_it));
    mem_needed += sizeof(str_t *);
  }

  /* alloc... */
  *elements = (const str_t **) malloc(mem_needed);
  if (!*elements)
    return errno;

  /* setup... */
  element = *elements;
  data = (char *) (*elements + *len);

  /* copy... */
  for (traverse_split_iterator_init(path, &it), element_number = 0;
       !traverse_split_iterator_ATEND(&it);
       traverse_split_iterator_incr(&it), ++element_number) {
    if (element_number == 2
	&& traverse_split_iterator_LEN(&it) == 1
	&& traverse_split_iterator_DATA(&it)[0] == '~') {
      *(element++) = user_str;
      *(element++) = username;
    } else {
      *element = (const str_t *) data;
      str_alloc(traverse_split_iterator_LEN(&it),
		traverse_split_iterator_DATA(&it),
		(str_t **) & data);
      ++element;
      data += str_MEM_NEEDED(traverse_split_iterator_LEN(&it));
    }
  }

  for (append_it = append, append_end = append + append_len;
       append_it < append_end;
       append_it++) {
    *element = (const str_t *) data;
    str_copy(*append_it, (str_t **) & data);
    data += str_MEM_NEEDED(str_LEN(*append_it));
    ++element;
  }

  /* done. */
  return OK;
}



static result_t
create_empty_dataset(dataset_key_t key,
		     stamp_t * stamp,
		     dataset_t ** dest)
{
  result_t        result;

  char            buf[100];	/* more than enough rope */

  str_t          *datapath = (str_t *) & buf;

  dataset_create_state_t create_state;

  I(stamp);
  I(dest);

  /* TODO: fix this horrible hack */
  sprintf(str_DATA(datapath), "ds%u", key);
  str_SET_LEN(datapath, strlen(str_DATA(datapath)));

  *dest = NULL;

  result = dataset_alloc(stamp,
			 datapath,
			 TRUE,
			 NULL, NULL,
			 dest,
			 &create_state);
  if (result != OK)
    goto cleanup;

  result = dataset_complete(dest, &create_state);
  if (result != OK)
    goto cleanup;

  return OK;

cleanup:
  if (*dest)
    dataset_free(dest);
  return result;
}

inline static result_t
add_attr(const str_t * attr_name,
	 const str_t * value,
	 entry_t ** new_entry,
	 entry_create_state_t * create_state)
{
  result_t        result;

  result = entry_add_attribute(attr_name,
			       new_entry,
			       ATTR_SINGLE,
			       create_state);
  if (result != OK)
    return result;

  return entry_add_attr_value(value,
			      new_entry,
			      create_state);
}

inline static result_t
check_add_required_attr(const attr_t * attr,
			const str_t * attr_name,
			const str_t * value,
			entry_t ** new_entry,
			entry_create_state_t * create_state,
			bool_t * added,
			bool_t * skip)
{
  result_t        result;

  compare_result_t compare_result;

  compare_result = str_compare(attr_NAME(attr), attr_name);
  if (compare_result == COMPARE_EQUAL) {
    *skip = TRUE;
  } else if (compare_result == COMPARE_GREATER) {
    result = add_attr(attr_name, value, new_entry, create_state);

    if (result != OK)
      return result;

    *added = TRUE;
  }
  return OK;
}

static result_t
traverse_create_dataset(entry_ref_t * entry_ref,
			dataset_db_t * entry_db,
			trx_t * trx,
			dataset_key_t * dataset_key,
			respcode_t ** respcode)
{

  /*
   * Okay!  We have to create an empty dataset_t within the entry_db, and
   * then go back and add a link to it for the parent, checking the parent's
   * dataset_data to see if the entry exists, and possibly making it exist
   * and checking its subdataset attribute.
   */

  /*
   * This is one, long, ugly, nasty, ugly, long, function; I don't see a
   * really good way to chop it down, though.  Hang on for the ride...
   */

  result_t        result;

  dataset_t      *dataset = NULL;

  dataset_ref_t  *parent_dataset_ref;

  dataset_data_t *parent_dataset_data;

  entry_t        *entry,
                 *new_entry;

  entry_iterator_t entry_it;

  entry_create_state_t create_state;

  const attr_t   *attr;

  bool_t          added_subdataset,
                  added_entry,
                  added_modtime;

  strlist_iterator_t strlist_it;

  str_t          *modtime = NULL;

  I(entry_ref);
  I(entry_db);
  I(trx);
  I(trx_IS_RW(trx));
  I(respcode);

  result = trx_str(trx, &modtime);
  if (result != OK)
    return result;

  parent_dataset_ref = entry_ref_PARENT(entry_ref);
  I(parent_dataset_ref);

  result = dataset_ref_get_dataset_data(parent_dataset_ref,
					&parent_dataset_data);
  if (result != OK)
    goto cleanup_modtime;

  result = dataset_data_get_entry(parent_dataset_data,
				  trx,
				  entry_ref_NAME(entry_ref),
				  &entry);
  if (result != OK)
    goto cleanup_parent_data;

  /* Try to find the "subdataset" attribute... */
  for (entry_ITER_BEGIN(entry, &entry_it);
       !entry_ITER_ATEND(&entry_it);
       entry_ITER_INCR(&entry_it)) {
    attr = entry_ATTR(entry_ITER_VALUE(&entry_it));
    if (str_equal(attr_NAME(attr), subdataset_str) == EQ_IS_EQUAL) {
      /* We got it.  Is it a redirect? */
      if ((attr_STATE(attr) == ATTR_SINGLE
	   && (str_equal_cstr(strlist_NTH_VALUE(attr_VALUE(attr), 0), ".")
	       == EQ_NOT_EQUAL))
	  || (attr_STATE(attr) == ATTR_MULTI
	      && 0 < strlist_N_VALUES(attr_VALUE(attr)))) {
	/* ooooo -- time to do a REFER! */
	result = respcode_refer_alloc(attr_VALUE(attr),
				      respcode);
	if (result == OK)
	  result = ACAP_ERR_REDIRECT;
	goto cleanup_entry;
      }
    }
  }

  result = dataset_db_store(entry_db, trx,
			    dataset_key, NULL);
  if (result != OK)
    goto cleanup_entry;

  result = create_empty_dataset(*dataset_key,
				trx_STAMP(trx),
				&dataset);

  if (result != OK)
    goto cleanup_db;

  result = dataset_db_replace(entry_db, trx, *dataset_key, dataset);
  if (result != OK)
    goto cleanup_dataset;

  dataset_free(&dataset);
  dataset = NULL;

  /* Now, we just need to add this entry to the parent... */
  result = dataset_db_replace_entry(dataset_ref_DB(parent_dataset_ref),
				    trx,
				    dataset_ref_KEY(parent_dataset_ref),
				    entry_ref_NAME(entry_ref),
				    *dataset_key);
  if (result != OK)
    goto cleanup_db;

  /* And set it in the entry_t */
  result = entry_alloc(trx_STAMP(trx),
		       entry_FLAGS(entry) | ENTRY_FLAG_VISIBLE,
		       *dataset_key,
		       &new_entry,
		       &create_state);
  if (result != OK)
    goto cleanup_db;

  added_subdataset = FALSE;
  added_entry = FALSE;
  added_modtime = FALSE;
  for (entry_ITER_BEGIN(entry, &entry_it);
       !entry_ITER_ATEND(&entry_it);
       entry_ITER_INCR(&entry_it)) {
    bool_t          skip = FALSE;

    attr = entry_ATTR(entry_ITER_VALUE(&entry_it));

    if (!added_entry) {
      result = check_add_required_attr(attr,
				       entry_str,
				       entry_ref_NAME(entry_ref),
				       &new_entry,
				       &create_state,
				       &added_entry,
				       &skip);
      if (result != OK)
	goto cleanup_new_entry;
    }
    if (added_entry && !added_modtime) {
      result = check_add_required_attr(attr,
				       modtime_str,
				       modtime,
				       &new_entry,
				       &create_state,
				       &added_modtime,
				       &skip);
      if (result != OK)
	goto cleanup_new_entry;
    }
    if (added_modtime && !added_subdataset) {
      result = check_add_required_attr(attr,
				       subdataset_str,
				       subdataset_dot,
				       &new_entry,
				       &create_state,
				       &added_subdataset,
				       &skip);
      if (result != OK)
	goto cleanup_new_entry;
    }
    if (skip == TRUE)
      continue;

    result = entry_add_attribute(attr_NAME(attr),
				 &new_entry,
				 attr_STATE(attr),
				 &create_state);
    if (result != OK)
      goto cleanup_new_entry;
    for (strlist_ITER_BEGIN(attr_ACL(attr), &strlist_it);
	 !strlist_ITER_ATEND(&strlist_it);
	 strlist_ITER_INCR(&strlist_it)) {
      result = entry_add_attr_acl(strlist_ITER_VALUE(&strlist_it),
				  &new_entry,
				  &create_state);
      if (result != OK)
	goto cleanup_new_entry;
    }

    for (strlist_ITER_BEGIN(attr_VALUE(attr), &strlist_it);
	 !strlist_ITER_ATEND(&strlist_it);
	 strlist_ITER_INCR(&strlist_it)) {
      result = entry_add_attr_value(strlist_ITER_VALUE(&strlist_it),
				    &new_entry,
				    &create_state);
      if (result != OK)
	goto cleanup_new_entry;
    }
    if (attr_HAS_CTYPE(attr)) {
      result = entry_add_attr_ctype(attr_CTYPE(attr),
				    &new_entry,
				    &create_state);
      if (result != OK)
	goto cleanup_new_entry;
    }
  }
  if (!added_entry) {
    result = add_attr(entry_str, entry_ref_NAME(entry_ref),
		      &new_entry, &create_state);
    if (result != OK)
      goto cleanup_new_entry;
  }
  if (!added_modtime) {
    result = add_attr(modtime_str, modtime,
		      &new_entry, &create_state);
    if (result != OK)
      goto cleanup_new_entry;
  }
  if (!added_subdataset) {
    result = add_attr(subdataset_str, subdataset_dot,
		      &new_entry, &create_state);
    if (result != OK)
      goto cleanup_new_entry;
  }
  result = entry_complete(&new_entry, &create_state);
  if (result != OK)
    goto cleanup_new_entry;

  result = dataset_data_replace_entry(parent_dataset_data,
				      trx,
				      entry_ref_NAME(entry_ref),
				      entry,
				      entry_ref_NAME(entry_ref),
				      new_entry);
  if (result != OK)
    goto cleanup_new_entry;

  /*
   * The dataset_data is responsible for cleaning up new_entry at this point.
   */
  new_entry = NULL;

  result = trx_lock_entry(trx, entry_ref);
  if (result != OK)
    goto cleanup_new_entry;

  I(entry_ref_WRITER(entry_ref));
  I(entry_ref_history_DBKEY(entry_ref_WRITER(entry_ref))
    == UNDETERMINED_DATASET_KEY);
  I(stamp_EQ(entry_ref_history_STAMP(entry_ref_WRITER(entry_ref)),
	     trx_STAMP(trx)));

  entry_ref_history_SET_DBKEY(entry_ref_WRITER(entry_ref),
			      *dataset_key);

  goto cleanup_entry;		/* don't unwind dataset_db changes */

cleanup_new_entry:
  if (new_entry)
    entry_free(&new_entry);

cleanup_dataset:
  if (dataset)
    dataset_free(&dataset);

cleanup_db:
  dataset_db_remove(entry_db, trx, *dataset_key);

cleanup_entry:
  dataset_data_done_with_entry(parent_dataset_data, &entry);

cleanup_parent_data:
  dataset_data_unref(&parent_dataset_data);

cleanup_modtime:
  str_free(&modtime);

  return result;
}

/* This actually looks up a dataset for an entry_ref. */
static result_t
traverse_lookup_dataset(entry_ref_t * entry_ref,
			entry_ref_history_t * history,
			trx_t * trx,
			traverse_create_t create,
			dataset_ref_t ** dest,
			respcode_t ** respcode)
{
  result_t        result = OK;

  dataset_db_t   *db;

  const str_t    *dbpath;

  dataset_key_t   key;

  I(entry_ref);
  I(history);
  I(trx);
  I(dest);
  I(respcode);

  *dest = NULL;

  if (entry_ref_history_DBKEY(history) != UNDETERMINED_DATASET_KEY) {
    key = entry_ref_history_DBKEY(history);
    dbpath = entry_ref_history_DBPATH(history);
  } else if (entry_ref_WRITER(entry_ref)
	     && (entry_ref_history_DBKEY(entry_ref_WRITER(entry_ref))
		 != UNDETERMINED_DATASET_KEY)) {
    key = entry_ref_history_DBKEY(entry_ref_WRITER(entry_ref));
    dbpath = entry_ref_history_DBPATH(entry_ref_WRITER(entry_ref));
  } else {
    key = UNDETERMINED_DATASET_KEY;
    dbpath = NULL;
  }

  if (dbpath) {
    result = dataset_db_fac_lookup(dbpath, &db);
  } else {
    I(entry_ref_PARENT(entry_ref));
    db = dataset_ref_DB(entry_ref_PARENT(entry_ref));
    dataset_db_ref(db);
  }

  if (result != OK)
    goto cleanup;

  if (key == UNDETERMINED_DATASET_KEY) {
    if (create == TRAVERSE_NO_CREATE_DATASET
	|| !trx_IS_RW(trx))
      result = ACAP_ERR_NO_SUCH_DATASET;
    else {
      VL(("Creating dataset: path components are:\n"));
      result = traverse_create_dataset(entry_ref, db, trx, &key, respcode);
    }
  }
  if (result != OK)
    goto cleanup_db_ref;

  I(key != UNDETERMINED_DATASET_KEY);
  result = dataset_db_lookup(db, trx, key, dest);

  /* Special case for creating the root dataset */
  if (result == ACAP_ERR_NO_SUCH_DATASET
      && key == ROOT_DATASET_KEY
      && entry_ref == root_entry
      && trx_IS_RW(trx)) {
    dataset_t      *dataset;

    result = create_empty_dataset(key,
				  trx_STAMP(trx),
				  &dataset);
    if (result != OK)
      goto cleanup_db_ref;

    result = dataset_db_replace(db,
				trx,
				key,
				dataset);
    dataset_free(&dataset);

    if (result != OK)
      goto cleanup_db_ref;

    /* ... and *now*, this ought to succeed. */
    result = dataset_db_lookup(db, trx, key, dest);
  }
cleanup_db_ref:
  dataset_db_unref(&db);	/* either way, drop the ref... */

cleanup:
  return result;
}

static result_t
traverse_get_history(entry_ref_t * entry_ref,
		     stamp_t * stamp,
		     entry_ref_history_t ** history)
{
  entry_ref_history_iter_t it;

  I(entry_ref);
  I(history);

  pthread_mutex_lock(entry_ref_MUTEX(entry_ref));

  if (entry_ref_WRITER(entry_ref)
      && entry_ref_history_IS_YOUNGER(entry_ref_WRITER(entry_ref), stamp)) {
    /*
     * There's a writer, and it's younger than we are. It's being nice and
     * letting us run, but we want to play with some data that's in flux, so
     * we need to wait for it to finish.
     */
    pthread_cond_wait(entry_ref_COND(entry_ref),
		      entry_ref_MUTEX(entry_ref));
    I(!entry_ref_WRITER(entry_ref));
  }
  /*
   * Now, we need to walk down the entry_ref's history chain to get the
   * correct version of the entry.
   */
  for (entry_ref_HIST_ITER_BEGIN(entry_ref, &it);
       !entry_ref_HIST_ITER_ATEND(&it);
       entry_ref_HIST_ITER_INCR(&it)) {
    *history = entry_ref_HIST_ITER_VALUE(&it);
    if (stamp_LT(entry_ref_history_STAMP(*history), stamp))
      break;
  }

  I(!entry_ref_HIST_ITER_ATEND(&it));	/* we should've found *something* */

  pthread_mutex_unlock(entry_ref_MUTEX(entry_ref));

  return OK;
}

result_t
traverse_get_dataset(entry_ref_t * entry_ref,
		     trx_t * trx,
		     traverse_create_t create,
		     dataset_ref_t ** dest,
		     respcode_t ** respcode)
{
  result_t        result;

  entry_ref_history_t *history;

  I(entry_ref);
  I(trx);
  I(dest);
  I(respcode);

  *dest = NULL;
  result = traverse_get_history(entry_ref, trx_STAMP(trx), &history);
  if (result != OK)
    return result;

  return traverse_lookup_dataset(entry_ref, history, trx, create,
				 dest, respcode);
}

result_t
traverse_read(const str_t * path,
	      client_t * client,/* for ~ expansion */
	      trx_t * trx,
	      unsigned *inherit_depth,
	      dataset_ref_t ** dest,
	      str_t ** inherit,
	      respcode_t ** respcode)
{
  result_t        result = OK;

  entry_ref_t    *entry_ref = NULL;

  dataset_ref_t  *dataset_ref = NULL;

  unsigned        n_elements;

  const str_t   **elements,
                **begin,
                **it,
                **end;

  entry_ref_history_t *history;

  I(path);
  I(dest);
  I(str_LEN(path));
  I(str_DATA(path)[0] == '/');

  *dest = NULL;
  if (inherit) {
    *inherit = NULL;
    if (inherit_depth && ACAP_INHERITANCE_DEPTH <= *inherit_depth)
      inherit = NULL;
  }
  VL(("Read-traversing path %s\n", str_DATA(path)));

  /* Find the components... */
  elements = NULL;
  result = traverse_split_path(path, client_USERNAME(client), 0, NULL,
			       &n_elements, &elements);
  if (result != OK)
    return result;

  I(n_elements);

  result = get_root_entry_ref(&entry_ref);
  if (result != OK)
    goto cleanup;

  /* And traverse. */
  begin = elements + 1;
  it = begin;
  end = elements + n_elements;

  while (it <= end) {
    result = traverse_get_history(entry_ref,
				  trx_STAMP(trx),
				  &history);
    if (result != OK)
      goto cleanup;

    if (entry_ref_history_DBKEY(history) == UNDETERMINED_DATASET_KEY) {
      if (inherit_depth
	  && *inherit_depth + 1 < ACAP_INHERITANCE_DEPTH
	  && *inherit) {
	/*
	 * We want to use inheritance to find the subdataset, if possible...
	 */
	unsigned        new_n_elements;

	const str_t   **new_elements;

	result = traverse_split_path(*inherit,
				     client_USERNAME(client),
				     end - it + 1,
				     it - 1,
				     &new_n_elements,
				     &new_elements);
	if (result != OK)
	  goto cleanup;

	free(elements);
	elements = new_elements;
	if (!new_n_elements) {
	  /* This shouldn't ever happen, but... */
	  result = ACAP_ERR_NO_SUCH_DATASET;
	  traverse_set_noexist(begin, it, respcode);
	  goto cleanup;
	}
	++*inherit_depth;
	dataset_ref_unref(&dataset_ref);
	entry_ref_unref(&entry_ref);
	result = get_root_entry_ref(&entry_ref);
	if (result != OK)
	  goto cleanup;
	begin = new_elements + 1;
	it = begin;
	end = new_elements + new_n_elements;
	n_elements = new_n_elements;
	str_free(inherit);
	*inherit = NULL;
	if (ACAP_INHERITANCE_DEPTH <= *inherit_depth)
	  inherit = NULL;
	continue;
      } else {
	/* No inheritance is possible */
	result = ACAP_ERR_NO_SUCH_DATASET;
	traverse_set_noexist(begin, it, respcode);
	goto cleanup;
      }
    }
    I(entry_ref_history_DBKEY(history) != UNDETERMINED_DATASET_KEY);

    result = traverse_lookup_dataset(entry_ref,
				     history,
				     trx,
				     TRAVERSE_NO_CREATE_DATASET,
				     &dataset_ref, respcode);
    if (result == ACAP_ERR_NO_SUCH_DATASET)
      traverse_set_noexist(begin, it, respcode);

    if (result != OK)
      goto cleanup;

    /* And we can get rid of the entry... */
    entry_ref_unref(&entry_ref);

    /* Time to handle the inheritance... */
    if (inherit) {
      dataset_t      *dataset;

      str_t          *new_inherit;

      result = dataset_ref_get_dataset(dataset_ref, trx, &dataset);
      if (result != OK)
	goto cleanup;
      if (dataset_DEFAULT_INHERIT(dataset)) {
	/*
	 * This dataset picks up its inheritance from its parent + name. If
	 * we don't already have a path, we don't do anything (we'd just be
	 * recomputing the dataset's path).
	 */
	if (*inherit) {
	  /* Append to existing path... */
	  result = str_append(&new_inherit,
			      STR_APPEND_STR_T, *inherit,
			      STR_APPEND_CHAR, '/',
			      STR_APPEND_STR_T, *it,
			      STR_APPEND_DONE);
	} else {
	  new_inherit = NULL;
	  result = OK;
	}
      } else {
	/*
	 * Otherwise, our new inheritance path is the same as the dataset's
	 * -- if any.
	 */
	if (dataset_INHERIT(dataset))
	  result = str_copy(dataset_INHERIT(dataset), &new_inherit);
	else {
	  new_inherit = NULL;
	  result = OK;
	}
      }
      if (result != OK)
	goto cleanup;
      if (*inherit)
	str_free(inherit);
      *inherit = new_inherit;
      VL(("Now inheriting from %s\n",
	  *inherit ? str_DATA(*inherit) : "(null)"));
    }
    if (it == end)
      break;			/* we're done */

    VL(("Traversing element %s\n", str_DATA(*it)));

    result = dataset_ref_lookup_entry(dataset_ref, trx, *it, &entry_ref);
    if (result != OK)
      goto cleanup;

    VL(("Got entry_ref %p\n", entry_ref));

    result = dataset_ref_unref(&dataset_ref);
    if (result != OK)
      goto cleanup;

    it++;
  }

  /* Nothing really magical; we've got the dataset */
  I(dataset_ref);
  *dest = dataset_ref;
  dataset_ref = NULL;

cleanup:
  if (entry_ref)
    entry_ref_unref(&entry_ref);

  if (dataset_ref)
    dataset_ref_unref(&dataset_ref);

  if (elements)
    free(elements);

  return result;
}

result_t
traverse_open(dataset_key_t key,
	      trx_t *trx,
	      dataset_ref_t ** dest)
{
  /* We're given a dataset_key; we want to return the corresponding
   * dataset_ref_t. */
  result_t result;
  dataset_db_t *dataset_db = NULL;

  I(key != UNDETERMINED_DATASET_KEY);
  I(dest);

  *dest = NULL;

  /* XXX If/when we go to a multiple master system, we're going to
   * have to do something a little smarter here. */

  result = dataset_db_fac_lookup(root_db_name_str, &dataset_db);
  if (result != OK)
    goto cleanup;
  
  result = dataset_db_lookup(dataset_db, trx, key, dest);
  if (result != OK)
    goto cleanup;
  
 cleanup:
  if (dataset_db)
    dataset_db_unref(&dataset_db);

  return result;
}

result_t
traverse_write(const str_t * path,
	       client_t * client,	/* for ~ expansion */
	       trx_t * trx,
	       traverse_create_t create,
	       entry_ref_t ** dest,
	       respcode_t ** respcode)
{
  result_t        result = OK;

  entry_ref_t    *entry_ref = NULL;

  entry_ref_history_t *history = NULL;

  dataset_ref_t  *dataset = NULL;

  unsigned        n_elements;

  const str_t   **elements,
                **begin,
                **it,
                **end;

  I(path);
  I(dest);
  I(str_LEN(path));
  I(str_DATA(path)[0] == '/');

  *dest = NULL;

  VL(("Traversing path %s\n", str_DATA(path)));

  /* Find the components... */
  elements = NULL;
  result = traverse_split_path(path, client_USERNAME(client), 0, NULL,
			       &n_elements, &elements);
  if (result != OK)
    return result;

  I(n_elements);

  /* And traverse. */
  begin = elements + 1;
  end = elements + n_elements;
  it = begin;

  result = get_root_entry_ref(&entry_ref);
  if (result != OK)
    goto cleanup;

  history = entry_ref_HIST(entry_ref);
  I(history);

  while (it < end) {
    VL(("Traversing element %s\n", str_DATA(*it)));

    result = traverse_lookup_dataset(entry_ref,
				     history,
				     trx,
				     create,
				     &dataset,
				     respcode);
    if (result == ACAP_ERR_NO_SUCH_DATASET)
      traverse_set_noexist(begin, it, respcode);

    if (result != OK)
      goto cleanup;

    result = entry_ref_unref(&entry_ref);
    if (result != OK)
      goto cleanup;

    result = dataset_ref_lookup_entry(dataset, trx,
				      *it, &entry_ref);
    if (result != OK)
      goto cleanup;

    result = dataset_ref_unref(&dataset);
    if (result != OK)
      goto cleanup;

    history = entry_ref_HIST(entry_ref);
    I(history);

    it++;
  }

  I(entry_ref);

  result = trx_lock_entry(trx, entry_ref);
  if (result != OK)
    goto cleanup;

  I(stamp_EQ(entry_ref_history_STAMP(entry_ref_WRITER(entry_ref)),
	     trx_STAMP(trx)));

  *dest = entry_ref;

  entry_ref = NULL;

cleanup:
  if (entry_ref)
    entry_ref_unref(&entry_ref);

  if (dataset)
    dataset_ref_unref(&dataset);

  if (elements)
    free(elements);

  return result;
}
