/* sock.c - socket implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define SOCK_C
#include <config.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>

#ifdef HAVE_SYS_FILIO_H
#include <sys/filio.h>
#endif

#ifdef HAVE_STROPTS_H
#include <stropts.h>
#endif

#include <termios.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <com_err.h>
#include <string.h>
#include "io.h"
#include "sock.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_daemon_io)

struct sock_s {
  int             fd;
  sock_handler_t *handler;
  void           *rock;
  short           local_port;
};

#define sock_FD(s) ((s)->fd)

static void
sock_io_handler(int fd, sock_t * sock)
{
  while (1) {
    result_t result;

    struct sockaddr_in sin;

    int             conn_fd;

    int             flag;

    int             len = sizeof(sin);

    I(sock);
    I(0 <= fd);

    VL(("Accepting a connection\n"));
    conn_fd = accept(fd, (struct sockaddr *) & sin, &len);
    if (conn_fd < 0)
      return;

    VL(("Connection is fd %d, from %s\n", fd, inet_ntoa(sin.sin_addr)));

    fcntl(conn_fd, F_SETOWN, getpid());

    /* Make sure we're using blocking IO... */
    if ((flag = fcntl(conn_fd, F_GETFL)) < 0) {
      close(conn_fd);
      continue;
    }
    flag &= ~O_NONBLOCK;

#ifdef O_ASYNC			/* use O_ASYNC to get signals */
    flag |= O_ASYNC;
#endif				/* O_ASYNC */

    if (fcntl(conn_fd, F_SETFL, flag) < 0) {
      close(conn_fd);
      continue;
    }

#ifdef I_SETSIG			/* use I_SETSIG to get signals */
    if (ioctl(conn_fd, I_SETSIG, S_INPUT) < 0) {
      close(conn_fd);
      continue;
    }
#endif				/* I_SETSIG */

    /* And allocate a connection */
    I(sock->handler);
    result = (sock->handler) (&sin, conn_fd, sock->rock);
    if (result) {
      com_err(program_name, result,
	      "establishing a connection from %s on port %d",
	      inet_ntoa(sin.sin_addr), sock->local_port);
      close(conn_fd);
      continue;
    }
  }
}

result_t
sock_open(const char *port_name,
	  short default_port,
	  sock_handler_t * handler,
	  void *rock,
	  sock_t ** sock)
{
  result_t        result;

  int             flag;

  short           port_number = 0;

  struct sockaddr_in sin;

  I(port_name || default_port);
  I(handler);
  I(sock);
  I(!*sock);

  *sock = malloc(sizeof(sock_t));
  if (!*sock)
    return errno;

  (*sock)->fd = -1;
  (*sock)->handler = handler;
  (*sock)->rock = rock;

  if (port_name) {
    /* See if the port works as a number... */
    port_number = htons(atoi(port_name));
    if (!port_number) {
      /* No?  Bummer.  Try getservbyname... */
      struct servent *se = NULL;

#ifdef HAVE_GETSERVBYNAME_R
      struct servent  sent;

      char            buf[GETSERVBYNAME_R_BUF_SZ];

#ifdef __linux
      /* XXX */
      getservbyname_r(port_name, "tcp", &sent,
		      buf, GETSERVBYNAME_R_BUF_SZ, &se);
#else				/* ! linux */
      se = getservbyname_r(port_name, "tcp", &sent,
			   buf, GETSERVBYNAME_R_BUF_SZ);
#endif				/* ! linux */

#else				/* ! HAVE_GETSERVBYNAME_R */
      se = getservbyname(port_name, "tcp");
#endif				/* ! HAVE_GETSERVBYNAME_R */

      if (se)
	port_number = se->s_port;
    }
  }
  if (!port_number)
    port_number = htons(default_port);

  (*sock)->local_port = ntohs(port_number);

  memset(&sin, 0L, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_port = port_number;
  sin.sin_addr.s_addr = INADDR_ANY;

  sock_FD(*sock) = socket(PF_INET, SOCK_STREAM, 0);
  if (sock_FD(*sock) < 0) {
    result = errno;
    goto cleanup_sock;
  }
  flag = 1;
  if (setsockopt(sock_FD(*sock), SOL_SOCKET, SO_REUSEADDR,
		 (const char *) &flag, sizeof(flag)) < 0) {
    result = errno;
    goto cleanup_sock;
  }
  if (bind(sock_FD(*sock),
	   (struct sockaddr *) & sin,
	   sizeof(struct sockaddr_in))) {
    result = errno;
    goto cleanup_sock;
  }
  fcntl(sock_FD(*sock), F_SETOWN, getpid());

  if ((flag = fcntl(sock_FD(*sock), F_GETFL)) < 0) {
    result = errno;
    goto cleanup_sock;
  }
  flag |= O_NONBLOCK

#ifdef O_ASYNC			/* use O_ASYNC to get signals */
    | O_ASYNC
#endif				/* O_ASYNC */

    ;
  if (fcntl(sock_FD(*sock), F_SETFL, flag) < 0) {
    result = errno;
    goto cleanup_sock;
  }

#ifdef I_SETSIG			/* use I_SETSIG to get signals */
  if (ioctl(sock_FD(*sock), I_SETSIG, S_INPUT) < 0) {
    result = errno;
    goto cleanup_sock;
  }
#endif				/* I_SETSIG */

  if (listen(sock_FD(*sock), SOMAXCONN)) {
    result = errno;
    goto cleanup_sock;
  }
  VL(("Listening for connection on fd %d\n", sock_FD(*sock)));

  result = io_add_handler(sock_FD(*sock),
			  (io_handler_t *) & sock_io_handler,
			  *sock);
  if (result)
    goto cleanup_sock;

  return OK;

cleanup_sock:
  if (0 <= (*sock)->fd)
    close((*sock)->fd);
  free(*sock);
  *sock = NULL;
  return result;
}

result_t
sock_close(sock_t ** sock)
{
  I(sock);
  I(*sock);

  io_remove_handler(sock_FD(*sock));

  if (0 <= (*sock)->fd)
    close((*sock)->fd);
  free(*sock);
  *sock = NULL;
  return OK;
}
