/* refcount.h - threaded refcount interface
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#ifndef REFCOUNT_H
#define REFCOUNT_H

#include <pthread.h>

/* Just a utility package for keeping track of reference counts.
 * It bothers me a little to use a pthread_mutex for this,
 * but I don't see a better system-independant way to do it.
 * It'd be nice to eventually special-case this for different
 * processors... */

typedef struct {
  pthread_mutex_t mutex;
  unsigned        refcount;
}               refcount_t;

#define refcount_init(r) ((r)->refcount = 0,				\
			  pthread_mutex_init(&(r)->mutex, NULL))

#define refcount_destroy(r) (pthread_mutex_destroy(&(r)->mutex))

/* Increments a refcount */
#define refcount_REF(r) (pthread_mutex_lock(&(r)->mutex),	\
			 ++(r)->refcount,			\
			 pthread_mutex_unlock(&(r)->mutex),	\
			 OK)

/* Decrements a refcount, running cleanup if refcount is now zero. */
#define refcount_UNREF(r, cleanup)		\
     (pthread_mutex_lock(&(r)->mutex),		\
      --(r)->refcount,				\
      (r)->refcount				\
      ? (pthread_mutex_unlock(&(r)->mutex), OK)	\
      : (pthread_mutex_unlock(&(r)->mutex),	\
	 (cleanup)))

/* Decrements a refcount, running cleanup if refcount is now zero,
 * or else running stillvalid. */
#define refcount_UNREF_ELSE(r, cleanup, stillvalid)		\
     (pthread_mutex_lock(&(r)->mutex),				\
      --(r)->refcount,						\
      (r)->refcount						\
      ? (pthread_mutex_unlock(&(r)->mutex), (stillvalid))	\
      : (pthread_mutex_unlock(&(r)->mutex),			\
	 (cleanup)))

/* Decrements a refcount, running check outside the mutex (and *not*
 * decrementing) if refcount is one. */
#define refcount_UNREF_CHECKLAST(r, check)	\
     (pthread_mutex_lock(&(r)->mutex),		\
      (r)->refcount == 1			\
      ? (pthread_mutex_unlock(&(r)->mutex),	\
	 (check))				\
      : (--(r)->refcount,			\
	 pthread_mutex_unlock(&(r)->mutex),	\
	 OK))
#endif				/* REFCOUNT_H */
