/* str.h - length-string interface
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#ifndef STR_H
#define STR_H

#include <sys/types.h>
#include <netinet/in.h>
#include <sys/param.h>

#ifdef HAVE_SYS_SYSMACROS_H
#include <sys/sysmacros.h>
#endif

/* This is a simple length-string system.  The only really interesting
 * things are that the length is always stored in 32bit network byte
 * order (because that way, they can be mmap()ed from a file created
 * on a different architecture, which will no doubt cause some poor
 * sysadmin to worship me someday), and that they may be arbitarily
 * moved around and they'll stay intact - either as part of a chunk of
 * memory being realloc()ed, or by memcpy()ing str_MEM_NEEDED bytes of
 * them.  */

/* To declare one of these statically, you need to put the length into
 * the first four octets in network byte order, followed by the data,
 * followed by zero padding to take the string out to a multiple of
 * ALIGNMENT_SIZE (which is typically eight); note that, if written
 * out as a quoted string, you can eliminate the last zero pad octet,
 * because the string will automatically have one appended.
 *
 * For instance, you could declare a str_t for "modtime" like so:
 * static const str_t *modtime_str = (str_t *) "\0\0\0\007modtime\0\0\0\0";
 */

/* We put the string data in memory just after the str_t, and append
 * at least one zero pad; it's not quite as clear as some other
 * approaches, but code accessing the data should be using the
 * str_DATA() macro anyway, and it's a lot simpler than trying to
 * declare the struct to hold a variable length string or having an
 * extra pointer to the string or something like that. */
typedef struct {
  uint32_t        len;		/* always in network byte order */
  /* Data immediately following... */
}               str_t;

#define str_LEN(s) (ntohl((s)->len))
#define str_DATA(s) ((char *)((s) + 1))

#define str_MEM_NEEDED(len) roundup((len) + sizeof(str_t) + 1, \
				    ALIGNMENT_SIZE)

/* Note: this should only be used if you *really* know
 * what you're doing, and you have a really good reason
 * for constructing the string in-place without
 * using the standard allocation functions. */
#define str_SET_LEN(s, l) ((s)->len = htonl((l)))

/* This may be used for placement construction, by setting
 * *str to point to enough memory for the string.  Note that the
 * memory *must* be aligned. */
extern          result_t
str_alloc(uint32_t length,
	  const char *buf,	/* NULL => no copy */
	  str_t ** str);

/* Likewise, this may be used for placement construction,
 * by setting *out to point to enough memory for the string.
 * OTOH, placement construction with str_copy is essentially the
 * same thing as memcpy(), and you don't have to check for errors
 * with memcpy(). */
extern result_t str_copy(const str_t * in, str_t ** out);

/* Yet again, *out != NULL => placement construction. */
extern result_t str_copy_cstr(const char *in, str_t ** out);

/* This is a fun one.  str_append() doesn't permit placement construction;
 * it always allocs on the heap.  It takes a number of str_t's and char *s,
 * and glues 'em all together into one str_t.  Precede each object with
 * a str_append_str_type_t identifier (so str_append knows what it's doing),
 * and terminate with STR_APPEND_DONE. */
typedef enum {
  STR_APPEND_DONE,
  STR_APPEND_STR_T,
  STR_APPEND_C_STRING,
  STR_APPEND_CHAR
}               str_append_str_type_t;

extern result_t str_append(str_t ** dest,...);

extern result_t 
str_join(const str_t * prefix,
	 const str_t * seperator,
	 const str_t * suffix,
	 const str_t * const * begin,
	 const str_t * const * end,
	 str_t ** dest);

extern result_t str_free(str_t ** str);

extern eq_result_t str_equal(const str_t * a, const str_t * b);

extern eq_result_t str_equal_cstr(const str_t * str, const char *cstr);

extern hashvalue_t str_hashvalue(const str_t * str);

extern compare_result_t str_compare(const str_t * a, const str_t * b);

typedef enum {
  STR_NOT_UTF8,
  STR_IS_UTF8
}               str_is_utf8_t;

extern str_is_utf8_t str_is_UTF8(const str_t * str);

#endif				/* STR_H */
